"""
Python bindings for XMMS2.
"""

# CImports
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from cpython.bytes cimport PyBytes_FromStringAndSize
cimport cython
from xmmsutils cimport from_unicode, to_unicode
from cxmmsvalue cimport *
from cxmmsclient cimport *
from xmmsvalue cimport *

# The following constants are meant for interpreting the return value of
# XMMS.playback_status ()
PLAYBACK_STATUS_STOP  = XMMS_PLAYBACK_STATUS_STOP
PLAYBACK_STATUS_PLAY  = XMMS_PLAYBACK_STATUS_PLAY
PLAYBACK_STATUS_PAUSE = XMMS_PLAYBACK_STATUS_PAUSE

PLAYLIST_CHANGED_ADD     = XMMS_PLAYLIST_CHANGED_ADD
PLAYLIST_CHANGED_INSERT  = XMMS_PLAYLIST_CHANGED_INSERT
PLAYLIST_CHANGED_SHUFFLE = XMMS_PLAYLIST_CHANGED_SHUFFLE
PLAYLIST_CHANGED_REMOVE  = XMMS_PLAYLIST_CHANGED_REMOVE
PLAYLIST_CHANGED_CLEAR   = XMMS_PLAYLIST_CHANGED_CLEAR
PLAYLIST_CHANGED_MOVE    = XMMS_PLAYLIST_CHANGED_MOVE
PLAYLIST_CHANGED_SORT    = XMMS_PLAYLIST_CHANGED_SORT
PLAYLIST_CHANGED_UPDATE  = XMMS_PLAYLIST_CHANGED_UPDATE

PLUGIN_TYPE_ALL    = XMMS_PLUGIN_TYPE_ALL
PLUGIN_TYPE_XFORM  = XMMS_PLUGIN_TYPE_XFORM
PLUGIN_TYPE_OUTPUT = XMMS_PLUGIN_TYPE_OUTPUT

MEDIALIB_ENTRY_STATUS_NEW           = XMMS_MEDIALIB_ENTRY_STATUS_NEW
MEDIALIB_ENTRY_STATUS_OK            = XMMS_MEDIALIB_ENTRY_STATUS_OK
MEDIALIB_ENTRY_STATUS_RESOLVING     = XMMS_MEDIALIB_ENTRY_STATUS_RESOLVING
MEDIALIB_ENTRY_STATUS_NOT_AVAILABLE = XMMS_MEDIALIB_ENTRY_STATUS_NOT_AVAILABLE
MEDIALIB_ENTRY_STATUS_REHASH        = XMMS_MEDIALIB_ENTRY_STATUS_REHASH

COLLECTION_CHANGED_ADD    = XMMS_COLLECTION_CHANGED_ADD
COLLECTION_CHANGED_UPDATE = XMMS_COLLECTION_CHANGED_UPDATE
COLLECTION_CHANGED_RENAME = XMMS_COLLECTION_CHANGED_RENAME
COLLECTION_CHANGED_REMOVE = XMMS_COLLECTION_CHANGED_REMOVE

PLAYBACK_SEEK_CUR = XMMS_PLAYBACK_SEEK_CUR
PLAYBACK_SEEK_SET = XMMS_PLAYBACK_SEEK_SET

COLLECTION_NS_COLLECTIONS = to_unicode(<char *>XMMS_COLLECTION_NS_COLLECTIONS)
COLLECTION_NS_PLAYLISTS = to_unicode(<char *>XMMS_COLLECTION_NS_PLAYLISTS)
COLLECTION_NS_ALL = to_unicode(<char *>XMMS_COLLECTION_NS_ALL)

ACTIVE_PLAYLIST = to_unicode(<char *>XMMS_ACTIVE_PLAYLIST)

#####################################################################

def deprecated(f):
	def deprecated_decorator(*a, **kw):
		from os import getenv
		if getenv('XMMS_PYTHON_SHOW_DEPRECATED'):
			from sys import stderr
			print >>stderr, "DEPRECATED: %s()" % f.__name__
		return f(*a, **kw)
	deprecated_decorator.__doc__ = f.__doc__
	deprecated_decorator.__name__ = f.__name__
	return deprecated_decorator

# Trick to not expose select in the module, but keep it accessible from builtins
# methods.
cdef object select
cdef _install_select():
	global select
	from select import select as _sel
	select = _sel
_install_select()

cdef char *check_namespace(object ns, bint can_be_all) except NULL:
	cdef char *n
	if ns == COLLECTION_NS_COLLECTIONS:
		n = <char *>XMMS_COLLECTION_NS_COLLECTIONS
	elif ns == COLLECTION_NS_PLAYLISTS:
		n = <char *>XMMS_COLLECTION_NS_PLAYLISTS
	elif can_be_all and ns == COLLECTION_NS_ALL:
		n = <char *>XMMS_COLLECTION_NS_ALL
	else:
		raise ValueError("Bad namespace")
	return n

cdef bint ResultNotifier(xmmsv_t *res, void *o) noexcept:
	cdef object xres
	xres = <object> o
	try:
		return xres()
	except:
		import traceback, sys
		exc = sys.exc_info()
		traceback.print_exception(*exc)
		return False

cdef void ResultDestroyNotifier(void *o) noexcept:
	cdef XmmsResult obj
	obj = <XmmsResult>o
	obj._cb = None
	obj.result_tracker.release_result(obj)

cdef class XmmsSourcePreference:
	"""
	Source preference manager
	"""
	#cdef object sources

	def __init__(self, sources = None):
		self.set(sources)

	def get(self):
		return self.sources

	def set(self, sources):
		if sources is None:
			self.sources = []
		else:
			self.sources = [enforce_unicode(s) for s in sources]

cdef class XmmsResultTracker:
	"""
	Class used by XmmsCore to track results that set a notifier.
	"""
	#cdef object results

	def __cinit__(self):
		self.results = []

	cdef track_result(self, XmmsResult r):
		self.results.append(r)

	cdef release_result(self, XmmsResult r):
		self.results.remove(r)

	cdef disconnect_all(self, bint unset_result):
		cdef XmmsResult r
		for r in self.results[:]: # Create a copy because r.disconnect() will remove the result from the list.
			r.disconnect()
			if unset_result: #This would become useless if results didn't xmmsc_ref() the connection.
				xmmsc_result_unref(r.res)
				r.res = NULL

cdef class XmmsResult:
	"""
	Class containing the results of some operation
	"""
	#cdef xmmsc_result_t *res
	#cdef object _cb
	#cdef bint _cb_issetup
	#cdef XmmsSourcePreference source_pref
	#cdef int ispropdict
	#cdef XmmsResultTracker result_tracker

	def __cinit__(self):
		self.res = NULL
		self.ispropdict = 0
		self._cb_issetup = False

	def __dealloc__(self):
		if self.res != NULL:
			xmmsc_result_unref(self.res)
			self.res = NULL

	cdef set_sourcepref(self, XmmsSourcePreference sourcepref):
		self.source_pref = sourcepref
	cdef set_result(self, xmmsc_result_t *res):
		self.res = res
	cdef set_callback(self, XmmsResultTracker rt, cb):
		"""
		Set a callback function for a result
		"""
		if cb is not None and not hasattr(cb, '__call__'):
			raise TypeError("Type '%s' is not callable" % cb.__class__.__name__)
		self._cb = cb
		if cb is not None and not self._cb_issetup:
			self.result_tracker = rt
			rt.track_result(self)
			xmmsc_result_notifier_set_full(self.res, ResultNotifier, <void *> self, ResultDestroyNotifier)
			self._cb_issetup = True

	cpdef disconnect(self):
		self._cb = None
		if self._cb_issetup and self.res != NULL:
			self._cb_issetup = False
			#xmmsc_result_ref(self.res) # Needed ? oO
			xmmsc_result_disconnect(self.res) #self.result_tracker.release_result() called in ResultDestroyNotifier()

	property callback:
		def __get__(self):
			return self._cb

	# XXX Kept for compatibility.
	@deprecated
	def _callback(self):
		"""
		:deprecated: Use `__call__` instead.
		"""
		try:
			ret = self()
		except:
			import traceback, sys
			traceback.print_exception(*sys.exc_info())
			return False
		return ret

	def __call__(self):
		cb = self.callback
		if cb is not None:
			ret = cb(self.xmmsvalue())
			return True if ret is None else ret
		return False

	cpdef wait(self):
		"""
		Wait for the result from the daemon.
		"""
		if self.res == NULL:
			raise RuntimeError("Uninitialized result")

		xmmsc_result_wait(self.res)

	cpdef is_error(self):
		"""

		:return: Whether the result represents an error or not.
		:rtype: bool
		"""
		return self.xmmsvalue().is_error()

	@deprecated
	def iserror(self):
		"""
		:deprecated: Use `XmmsResult.is_error` instead.
		"""
		return self.is_error()

	cpdef xmmsvalue(self):
		cdef XmmsValue obj
		cdef xmmsv_t *value
		if self.res == NULL:
			raise RuntimeError("Uninitialized result")
		value = xmmsc_result_get_value(self.res)
		obj = XmmsValue(self.source_pref.get())
		obj.set_value(value, self.ispropdict)
		return obj

	property xvalue:
		def __get__(self):
			return self.xmmsvalue()

	@deprecated
	def _value(self):
		"""
		:deprecated: Use `xmmsvalue` or the xvalue property.
		"""
		return self.xmmsvalue()

	cpdef value(self):
		return self.xmmsvalue().value()


cdef class XmmsVisResult(XmmsResult):
	#cdef XmmsValue _val
	#cdef VisResultCommand command
	#cdef xmmsc_connection_t *conn

	def __cinit__(self):
		self.command = VIS_RESULT_CMD_NONE
		self.conn = NULL

	def __dealloc__(self):
		if self.conn != NULL:
			xmmsc_unref(self.conn)
			self.conn = NULL

	cdef set_command(self, VisResultCommand cmd, xmmsc_connection_t *conn):
		self.command = cmd
		if self.conn != NULL:
			xmmsc_unref(self.conn)
			self.conn = NULL
		if conn != NULL:
			self.conn = xmmsc_ref(conn)

	cdef retrieve_error(self):
		cdef xmmsv_t *errval
		xval = XmmsResult.xmmsvalue(self)
		if xval.is_error():
			self._val = xval
		else:
			self._val = XmmsValue()
			errval = xmmsv_new_error("Failed to initialize visualization")
			self._val.set_value(errval)
			xmmsv_unref(errval)

	cdef _init_xmmsvalue(self):
		if self._val is None:
			if self.res == NULL:
				raise RuntimeError("Uninitialized result")
			hid = xmmsc_visualization_init_handle(self.res)
			if hid == -1:
				self.retrieve_error()
			else:
				self._val = XmmsValue(pyval = hid)
		return self._val

	cdef _start_xmmsvalue(self):
		if self._val is None:
			if self.res == NULL:
				raise RuntimeError("Uninitialized result")
			xval = XmmsResult.xmmsvalue(self)
			if xval.is_error():
				self._val = xval
			elif self.conn == NULL:
				raise RuntimeError("Internal connection reference not set")
			else:
				xmmsc_visualization_start_handle(self.conn, self.res)
				self._val = XmmsValue(pyval = None)
		return self._val

	cpdef xmmsvalue(self):
		if self.command == VIS_RESULT_CMD_INIT:
			return self._init_xmmsvalue()
		elif self.command == VIS_RESULT_CMD_START:
			return self._start_xmmsvalue()
		else:
			# XXX Raise an exception ?
			return XmmsResult.xmmsvalue(self)

cdef class XmmsVisChunk:
	#cdef short *data
	#cdef int sample_count

	def __cinit__(self):
		self.data = NULL
		self.sample_count = 0

	def __dealloc__(self):
		if self.data != NULL:
			PyMem_Free(self.data)

	cdef set_data(self, short *data, int sample_count):
		if self.data != NULL:
			PyMem_Free(self.data)
			self.sample_count = 0
		self.data = <short *>PyMem_Malloc(sizeof (short) * sample_count)
		if self.data == NULL:
			raise RuntimeError("Failed to initialize chunk data")
		for i in range(sample_count):
			self.data[i] = data[i]
		self.sample_count = sample_count

	def __len__(self):
		return self.sample_count

	cpdef get_buffer(self):
		"""
		Get the chunk buffer

		:return: chunk data as a string
		"""
		if self.data == NULL:
			raise RuntimeError("chunk data not initialized")
		return PyBytes_FromStringAndSize(<char *>self.data, sizeof (short) * self.sample_count)

	cpdef get_data(self):
		"""
		Get chunk data as a list.

		:return: A list of int
		"""
		if self.data == NULL:
			raise RuntimeError("chunk data not initialized")
		return [<int>self.data[i] for i in range(self.sample_count)]

class VisualizationError(Exception):
	pass


def _noop(*a, **kw):
	pass
def _setcb(f, fb=_noop):
	return hasattr(f, '__call__') and f or fb

class method_arg:
	TYPES = {
		'integer': XMMSV_TYPE_INT64,
		'i': XMMSV_TYPE_INT64,
		'float': XMMSV_TYPE_FLOAT,
		'f': XMMSV_TYPE_FLOAT,
		'string': XMMSV_TYPE_STRING,
		's': XMMSV_TYPE_STRING,
		'coll': XMMSV_TYPE_COLL,
		'c': XMMSV_TYPE_COLL,
		'list': XMMSV_TYPE_LIST,
		'l': XMMSV_TYPE_LIST,
		'dict': XMMSV_TYPE_DICT,
		'd': XMMSV_TYPE_DICT,
		}
	def __init__(self, name, type = '', doc = '', **kw):
		for key, value in kw.items():
			setattr(self, key, value)
		self.name = name
		self.type = type.lower()
		if type not in self.TYPES:
			raise TypeError("Unsupported argument type %s" % type)
		self._x_type = self.TYPES.get(type, XMMSV_TYPE_NONE)
		self.doc = doc

class method_varg:
	def __init__(self, doc = '', **kw):
		self.doc = doc
		for key, value in kw.items():
			setattr(self, key, value)

class service_method:
	def __init__(self, positional = None, named = None, name = None, doc = None):
		self.positional = positional or ()
		self.named = named or ()
		self.name = name
		self.doc = doc

	def __call__(self, f):
		f._xmms_service_entity = dict(
				entity = 'method',
				positional = self.positional,
				named = self.named,
				name = self.name or f.__name__,
				doc = self.doc or f.__doc__,
				)
		return f

cdef class service_broadcast:
	def __init__(self, name = None, doc = None):
		self.name = name
		self.doc = doc

	cdef bind(self, XmmsServiceNamespace namespace, name = None):
		cdef service_broadcast obj
		obj = service_broadcast(self.name or name, self.doc)
		obj.namespace = namespace
		return obj

	cpdef emit(self, value = None):
		if not self.namespace:
			raise RuntimeError("No namespace bound to broadcast")
		path = self.namespace.path + (self.name,)
		return self.namespace.xmms.sc_broadcast_emit(path, value)

	def __call__(self, value = None):
		return self.emit(value)

class service_constant:
	def __init__(self, value):
		self.value = value

cdef class XmmsServiceNamespace:
	namespace_path = ()

	property path:
		def __get__(self):
			if self.parent:
				path = self.parent.path
			else:
				path = ()
			return path + self.namespace_path

	def __cinit__(self, XmmsCore xmms, *a, **ka):
		cdef service_broadcast bc_obj
		self.parent = ka.get('parent', None)
		self.xmms = xmms
		self._xmms_service_constants = set()
		self._bound_methods = dict()

		for attr in dir(self):
			if attr[0] == '_':
				continue
			obj = getattr(self, attr)

			if isinstance(obj, service_constant):
				self._xmms_service_constants.add(attr)
				setattr(self, attr, obj.value)
				continue

			if isinstance(obj, service_broadcast):
				bc_obj = obj
				setattr(self, attr, bc_obj.bind(self, attr))
				continue

			try:
				if issubclass(obj, XmmsServiceNamespace):
					inst = obj(self.xmms, parent = self)
					inst.namespace_path = (attr,)

					setattr(self, attr, inst)
					inst._xmms_service_entity = dict(
							entity = 'namespace',
							name = attr,
							doc = inst.__doc__
							)
			except TypeError:
				pass

	cpdef _walk(self, namespace = None, constant = None, method = None, broadcast = None, path = (), other = _noop, depth = -1, deep = False):
		callbacks = dict(
				namespace = _setcb(namespace, other),
				constant = _setcb(constant, other),
				method = _setcb(method, other),
				broadcast = _setcb(broadcast, other),
				)

		for attr in dir(self):
			if attr[0] == '_':
				continue
			obj = getattr(self, attr)
			if attr in self._xmms_service_constants:
				entity = dict(
						entity = 'constant',
						name = attr,
						)
			elif isinstance(obj, service_broadcast):
				entity = dict(
						entity = 'broadcast',
						name = obj.name or attr,
						doc = obj.doc
						)
			else:
				entity = getattr(obj, '_xmms_service_entity', None)

			if not isinstance(entity, dict):
				continue

			entity_type = entity.get('entity', '')

			if entity_type not in callbacks:
				continue

			if not deep:
				callbacks[entity_type](path, obj, entity)
			if entity_type == 'namespace' and depth:
				obj._walk(path = path + (entity['name'] or attr,), depth = depth - 1, **callbacks)
			if deep:
				callbacks[entity_type](path, obj, entity)

	cpdef register_namespace(self, path, instance, infos):
		cdef xmmsc_sc_namespace_t *ns
		name = from_unicode(infos.get('name', ""))
		if not name: #root namespace already exists
			return
		doc = from_unicode(infos.get('doc', ""))
		ns = _namespace_get(self.xmms.conn, path, True)
		xmmsc_sc_namespace_new(ns, <char *>name, <char *>doc)

	cpdef register_constant(self, path, value, infos):
		cdef xmmsc_sc_namespace_t *ns
		cdef xmmsv_t *val
		name = from_unicode(infos.get('name', ""))
		if not name:
			return
		ns = _namespace_get(self.xmms.conn, path, True)
		val = create_native_value(value)
		xmmsc_sc_namespace_add_constant(ns, <char *>name, val)
		xmmsv_unref(val)

	cpdef register_method(self, path, meth, infos):
		cdef xmmsc_sc_namespace_t *ns
		cdef xmmsv_t *positional
		cdef xmmsv_t *named
		cdef xmmsv_t *x_arg
		cdef bint va_pos
		cdef bint va_named
		cdef xmmsv_t *a_default
		cdef xmmsv_type_t a_type

		positional = xmmsv_new_list()
		named = xmmsv_new_list()

		name = from_unicode(infos.get('name') or "")
		if not name:
			return
		doc = from_unicode(infos.get('doc') or "")
		for arg in infos.get('positional', ()):
			if isinstance(arg, method_varg):
				va_pos = True
				break
			if hasattr(arg, 'default'):
				a_default = create_native_value(arg.default)
			else:
				a_default = NULL
			a_type = arg._x_type
			a_name = from_unicode(arg.name)
			a_doc = from_unicode(arg.doc)
			x_arg = xmmsv_sc_argument_new(<char *>a_name, <char *>a_doc, a_type, a_default)
			xmmsv_list_append(positional, x_arg)
			xmmsv_unref(x_arg)

		for arg in infos.get('named', ()):
			if isinstance(arg, method_varg):
				va_named = True
				break
			if hasattr(arg, 'default'):
				a_default = create_native_value(arg.default)
			else:
				a_default = NULL
			a_type = arg._x_type
			a_name = from_unicode(arg.name)
			a_doc = from_unicode(arg.doc)
			x_arg = xmmsv_sc_argument_new(<char *>a_name, <char *>a_doc, a_type, a_default)
			xmmsv_list_append(named, x_arg)
			xmmsv_unref(x_arg)

		ns = _namespace_get(self.xmms.conn, path, True)

		# We need this to prevent bound function wrapper to be recycled on
		# return, leading to weird behaviour
		self._bound_methods[name] = meth

		xmmsc_sc_namespace_add_method(ns, service_method_proxy, <char *>name, <char *>doc, positional, named, va_pos, va_named, <void *>meth)
		xmmsv_unref(positional)
		xmmsv_unref(named)

	cpdef register_broadcast(self, path, bc, infos):
		cdef xmmsc_sc_namespace_t *ns
		name = from_unicode(infos.get('name') or "")
		if not name:
			return
		doc = from_unicode(infos.get('doc') or "")
		ns = _namespace_get(self.xmms.conn, path, True)
		xmmsc_sc_namespace_add_broadcast(ns, <char *>name, <char *>doc)

	cpdef register(self):
		if self.registered:
			return

		self.xmms.sc_init()

		_path = self.path
		self.register_namespace(_path[:-1], self, dict(
			name = _path and _path[-1] or "",
			doc = self.__doc__ or ""
			))

		self._walk( path = _path,
			namespace = self.register_namespace,
			constant = self.register_constant,
			method = self.register_method,
			broadcast = self.register_broadcast)

		self.xmms.c2c_ready()

		self.registered = True


cdef xmmsc_sc_namespace_t *_namespace_get(xmmsc_connection_t *c, path, bint create):
	cdef xmmsc_sc_namespace_t *ns
	cdef xmmsc_sc_namespace_t *child_ns

	ns = xmmsc_sc_namespace_root(c)

	for p in path:
		p = from_unicode(p)
		doc = from_unicode("")
		child_ns = xmmsc_sc_namespace_get(ns, <char *>p)
		if child_ns == NULL:
			if not create:
				break
			child_ns = xmmsc_sc_namespace_new(ns, <char *>p, <char *>doc)
		ns = child_ns

	return ns

cdef xmmsv_t *service_method_proxy(xmmsv_t *pargs, xmmsv_t *nargs, void *udata) noexcept:
	cdef object method
	cdef XmmsValue args
	cdef XmmsValue kargs
	cdef xmmsv_t *res

	method = <object> udata
	args = XmmsValue()
	args.set_value(pargs)
	kargs = XmmsValue()
	kargs.set_value(nargs)
	try:
		ret = method(*args.get_list(), **kargs.get_dict())
		res = create_native_value(ret)
	except:
		import traceback, sys
		exc = sys.exc_info()
		traceback.print_exception(*exc)
		s = from_unicode("%s"%exc[1])
		res = xmmsv_new_error(<char *>s)
	return res


cdef class client_broadcast:
	def __cinit__(self, _XmmsServiceClient parent, name, docstring):
		if not parent._async:
			raise RuntimeError("Broadcast is available only on asynchronous connections")
		self._xmms = parent._xmms
		self._clientid = parent._clientid
		self._path = parent._path + (name,)
		self.name = name
		self.docstring = docstring

	def subscribe(self, cb = None):
		return self._xmms.sc_broadcast_subscribe(self._clientid, self._path, cb = cb)

	def __call__(self, cb = None):
		return self.subscribe(cb = cb)

	def __repr__(self):
		return "<broadcast '%s' on client #%d>" % (".".join(self._path), self._clientid)


cdef class client_method:
	def __cinit__(self, _XmmsServiceClient parent, name, docstring, inspect, cb = None):
		self._parent = parent
		self._callback = cb
		self._path = parent._path + (name,)
		self._clientid = parent._clientid
		self.name = name
		self.docstring = docstring
		self.inspect = inspect

	def with_callback(self, cb):
		if not self._parent._async:
			raise NotImplementedError()
		return client_method(self._parent, self.name, self.docstring, self.inspect, cb)

	def call(self, args, kargs, cb = None):
		res = self._parent._xmms.sc_call(self._clientid, self._path, args, kargs, cb = self._parent._async and cb or None)

		if self._parent._async:
			return res

		res.wait()
		xvalue = res.xvalue
		if xvalue.is_error():
			raise xvalue.get_error()
		cxvalue = XmmsValueC2C(pyval = xvalue)
		payload = cxvalue.payload
		if payload is not None:
			if payload.is_error():
				raise payload.get_error()
			payload = payload.value()
		return payload

	def __call__(self, *args, **kargs):
		return self.call(args, kargs, cb = self._callback)

	def __repr__(self):
		return "<method '%s' on client #%d>" % (".".join(self._path), self._clientid)


cdef class _XmmsServiceClient:
	def __cinit__(self, xmms, int clientid, *a, **ka):
		cdef XmmsProxy _proxy

		if isinstance(xmms, XmmsCore):
			self._xmms = xmms
			self._async = ka.get('_async', True)
		elif isinstance(xmms, XmmsProxy):
			_proxy = xmms
			self._xmms = _proxy._get_xmms()
			self._async = False
		else:
			raise TypeError("Not a valid Xmms client instance")

		self._path = tuple(ka.get('path', ()))
		self._clientid = clientid

	def __repr__(self):
		return "<namespace '%s' on client #%d>" % (".".join(self._path), self._clientid)

class XmmsServiceClient(_XmmsServiceClient):
	def _update_api(self, xvalue, recursive = False, cb = None):
		cdef XmmsValueC2C cxvalue
		cxvalue = XmmsValueC2C(pyval = xvalue)

		if cxvalue.sender != self._clientid:
			# ignore bad sender
			return

		payload = cxvalue.payload
		if payload and payload.get_type() == XMMSV_TYPE_DICT:
			payload = payload.get_dict()
		else:
			return

		self.__doc__ = payload.get('docstring', '')

		try:
			for c, v in payload.get('constants', {}).items():
				if hasattr(self.__class__, c):
					continue
				setattr(self, c, v)
		except (AttributeError, TypeError):
			pass

		if self._async:
			# Expose broadcasts only for asynchronous connections.
			for bc in payload.get('broadcasts', []):
				try:
					name = bc['name']
					doc = bc.get('docstring')
				except (TypeError, AttributeError, KeyError):
					continue
				if hasattr(self.__class__, name):
					continue
				setattr(self, name, client_broadcast(self, name, doc))

		for m in payload.get('methods', []):
			try:
				name = m['name']
				doc = m.get('docstring')
			except (TypeError, AttributeError, KeyError):
				continue
			if hasattr(self.__class__, name):
				continue
			setattr(self, name, client_method(self, name, doc, m))

		for ns in payload.get('namespaces', []):
			if hasattr(self.__class__, ns):
				continue
			sc = XmmsServiceClient(self._xmms, self._clientid,
					path = self._path + (ns,),
					_async = self._async)
			setattr(self, ns, sc)
			if recursive:
				sc(recursive, cb)

	def _update_api_callback(self, cb = None, recursive = False):
		def _update_api(xvalue):
			self._update_api(xvalue, recursive, cb)
			if cb:
				cb(self, xvalue)
		return _update_api

	def __call__(self, recursive = False, cb = None):
		cdef XmmsResult res
		cb = self._update_api_callback(cb, recursive)
		res = self._xmms.sc_introspect_namespace(self._clientid, self._path, cb = self._async and cb or None)

		if self._async:
			return res

		res.wait()
		xvalue = res.xvalue
		if xvalue.is_error():
			raise xvalue.get_error()
		cb(xvalue)


cdef void python_need_out_fun(int i, void *obj) noexcept:
	cdef object o
	o = <object> obj
	o._needout_cb(i)

cdef void python_disconnect_fun(void *obj) noexcept:
	cdef object o
	o = <object> obj
	o._disconnect_cb()

cpdef userconfdir_get():
	"""
	Get the user configuration directory, where XMMS2 stores its
	user-specific configuration files. Clients may store their
	configuration under the 'clients' subdirectory. This varies from
	platform to platform so should always be retrieved at runtime.
	"""
	cdef char path[XMMS_PATH_MAX]
	if xmmsc_userconfdir_get (path, XMMS_PATH_MAX) == NULL:
		return None
	return path

def enforce_unicode(object o):
	if isinstance(o, unicode):
		return o
	else:
		try:
			return unicode(o, "UTF-8")
		except:
			return o

cdef object check_playlist(object pls, bint None_is_active):
	if pls is None and not None_is_active:
		raise TypeError("expected str, %s found" % pls.__class__.__name__)
	return from_unicode(pls or ACTIVE_PLAYLIST)

cdef class XmmsProxy:
	def __init__(self, xmms):
		self._xmms = xmms

	cdef XmmsCore _get_xmms(self):
		return self._xmms

cdef class XmmsCore:
	"""
	This is the class representing the XMMS2 client itself. The methods in
	this class may be used to control and interact with XMMS2.
	"""
	#cdef xmmsc_connection_t *conn
	#cdef int isconnected
	#cdef object disconnect_fun
	#cdef object needout_fun
	#cdef readonly XmmsSourcePreference source_preference
	#cdef XmmsResultTracker result_tracker
	#cdef readonly object clientname

	def __cinit__(self, *args, **kargs): #Trick to allow subclass with init arguments
		"""
		Initiates a connection to the XMMS2 daemon. All operations
		involving the daemon are done via this connection.
		"""
		if 'clientname' in kargs:
			clientname = kargs['clientname']
		elif len(args):
			clientname = args[0]
		else:
			clientname = None
		if not clientname:
			clientname = "UnnamedPythonClient"
		self.clientname = clientname
		default_sp = get_default_source_pref()
		default_sp.insert(0, enforce_unicode("client/"+clientname))
		self.source_preference = XmmsSourcePreference(default_sp)
		self.result_tracker = XmmsResultTracker() # Keep track of all results that set a notifier.
		self.new_connection()

	def __init__(self, clientname = None):
		pass

	def __dealloc__(self):
		if self.conn != NULL:
			xmmsc_unref(self.conn)
			self.conn = NULL

	def __del__(self):
		self.disconnect()

	cdef new_connection(self):
		cn = from_unicode(self.clientname)
		self.conn = xmmsc_init(<char *>cn)
		if self.conn == NULL:
			raise ValueError("Failed to initialize xmmsclient library! Probably due to broken name.")

	cpdef get_source_preference(self):
		return self.source_preference.get()
	cpdef set_source_preference(self, sources):
		self.source_preference.set(sources)

	cpdef _needout_cb(self, int i):
		if self.needout_fun:
			self.needout_fun(i)
	cpdef _disconnect_cb(self):
		if self.disconnect_fun:
			self.disconnect_fun(self)

	cpdef disconnect(self):
		# Disconnect all results.
		self.result_tracker.disconnect_all(True)
		if self.conn:
			xmmsc_unref(self.conn)
			self.conn = NULL
		self.new_connection()

	cpdef ioin(self):
		"""
		Read data from the daemon, when available.
		Note: This is a low level function that should only be used in
		certain circumstances. e.g. a custom event loop
		"""
		return xmmsc_io_in_handle(self.conn)

	cpdef ioout(self):
		"""
		Write data out to the daemon, when available. Note: This is a
		low level function that should only be used in certain
		circumstances. e.g. a custom event loop
		"""
		return xmmsc_io_out_handle(self.conn)

	cpdef want_ioout(self):
		"""
        Check if there's data to send to the server.
		"""
		return xmmsc_io_want_out(self.conn)

	cpdef set_need_out_fun(self, fun):
		"""
        Set a function to be called when there's data to be
        sent to the server.
		"""
		self.needout_fun = fun

	cpdef get_fd(self):
		"""
		Get the underlying file descriptor used for communication with
		the XMMS2 daemon. You can use this in a client to ensure that
		the IPC link is still active and safe to use.(e.g by calling
		select() or poll())

		:return: IPC file descriptor
		:rtype: int
		"""
		return xmmsc_io_fd_get(self.conn)

	cpdef connect(self, path = None, disconnect_func = None):
		"""
		Connect to the appropriate IPC path, for communication with the
		XMMS2 daemon. This path defaults to /tmp/xmms-ipc-<username> if
		not specified. Call this once you have instantiated the object:

		>>> import xmmsclient
		>>> xc = xmmsclient.Xmms('clientname')
		>>> xc.connect()

		You can provide a disconnect callback function to be activated
		when the daemon disconnects.(e.g. daemon quit) This function
		typically has to exit the main loop used by your application.
		For example, if using `loop`, your callback should call
		`exit_loop` at some point.
		"""
		if path:
			p = from_unicode(path)
			ret = xmmsc_connect(self.conn, <char *>p)
		else:
			ret = xmmsc_connect(self.conn, NULL)

		if not ret:
			raise IOError("Couldn't connect to the server")
		self.isconnected = 1

		self.disconnect_fun = disconnect_func
		xmmsc_disconnect_callback_set_full(self.conn, python_disconnect_fun, <void *>self, NULL)
		xmmsc_io_need_out_callback_set_full(self.conn, python_need_out_fun, <void *> self, NULL)

	def is_connected(self):
		return self.isconnected != 0

	cdef XmmsResult _create_result(self, cb, xmmsc_result_t *res, Cls):
		cdef XmmsResult ret
		if res == NULL:
			raise RuntimeError("xmmsc_result_t couldn't be allocated")
		ret = Cls()
		ret.set_sourcepref(self.source_preference)
		ret.set_result(res)
		if cb is not None:
			ret.set_callback(self.result_tracker, cb) # property that setup all.
		return ret

	cdef XmmsResult create_result(self, cb, xmmsc_result_t *res):
		return self._create_result(cb, res, XmmsResult)

	cdef XmmsResult create_vis_result(self, cb, xmmsc_result_t *res, VisResultCommand cmd):
		cdef XmmsVisResult vres = self._create_result(cb, res, XmmsVisResult)
		vres.set_command(cmd, self.conn)
		return vres


cdef class XmmsApi(XmmsCore):
	property client_id:
		def __get__(self):
			return self.c2c_get_own_id()

	cpdef int c2c_get_own_id(self):
		return xmmsc_c2c_get_own_id(self.conn)

	cpdef XmmsResult quit(self, cb = None):
		"""
		Tell the XMMS2 daemon to quit.

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_quit(self.conn))

	cpdef XmmsResult plugin_list(self, typ, cb = None):
		"""
		Get a list of loaded plugins from the server

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_main_list_plugins(self.conn, typ))

	cpdef XmmsResult playback_start(self, cb = None):
		"""
		Instruct the XMMS2 daemon to start playing the currently
		selected file from the playlist.

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_playback_start(self.conn))

	cpdef XmmsResult playback_stop(self, cb = None):
		"""
		Instruct the XMMS2 daemon to stop playing the file
		currently being played.

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_playback_stop(self.conn))

	cpdef XmmsResult playback_tickle(self, cb = None):
		"""
		Instruct the XMMS2 daemon to move on to the next playlist item.

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_playback_tickle(self.conn))

	cpdef XmmsResult playback_pause(self, cb = None):
		"""
		Instruct the XMMS2 daemon to pause playback.

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_playback_pause(self.conn))

	cpdef XmmsResult playback_current_id(self, cb = None):
		"""

		:return: The medialib id of the item currently selected.
		"""
		return self.create_result(cb, xmmsc_playback_current_id(self.conn))

	cpdef XmmsResult playback_seek_ms(self, int ms, xmms_playback_seek_mode_t whence = PLAYBACK_SEEK_SET, cb = None):
		"""
		Seek to a time position in the current file or stream in playback.

		:return: The result of the operation.
		"""
		if whence == PLAYBACK_SEEK_SET or whence == PLAYBACK_SEEK_CUR:
			return self.create_result(cb, xmmsc_playback_seek_ms(self.conn, ms, whence))
		else:
			raise ValueError("Bad whence parameter")

	cpdef XmmsResult playback_seek_samples(self, int samples, xmms_playback_seek_mode_t whence = PLAYBACK_SEEK_SET, cb = None):
		"""
		Seek to a number of samples in the current file or stream in playback.

		:return: The result of the operation.
		"""
		if whence == PLAYBACK_SEEK_SET or whence == PLAYBACK_SEEK_CUR:
			return self.create_result(cb, xmmsc_playback_seek_samples(self.conn, samples, whence))
		else:
			raise ValueError("Bad whence parameter")

	cpdef XmmsResult playback_status(self, cb = None):
		"""
		Get current playback status from XMMS2 daemon. This is
		essentially the more direct version of
		`broadcast_playback_status`. Possible return values are:
		`PLAYBACK_STATUS_STOP`, `PLAYBACK_STATUS_PLAY`,
		`PLAYBACK_STATUS_PAUSE`

		:return: Current playback status
		"""
		return self.create_result(cb, xmmsc_playback_status(self.conn))

	cpdef XmmsResult broadcast_playback_status(self, cb = None):
		"""
		Set a method to handle the playback status broadcast from the
		XMMS2 daemon.
		"""
		return self.create_result(cb, xmmsc_broadcast_playback_status(self.conn))

	cpdef XmmsResult broadcast_playback_current_id(self, cb = None):
		"""
		Set a method to handle the playback id broadcast from the
		XMMS2 daemon.
		"""
		return self.create_result(cb, xmmsc_broadcast_playback_current_id(self.conn))

	cpdef XmmsResult playback_playtime(self, cb = None):
		"""
		Return playtime on current file/stream. This is essentially a
		more direct version of `signal_playback_playtime`

		:return: The result of the operation (playtime in milliseconds).
		"""
		return self.create_result(cb, xmmsc_playback_playtime(self.conn))

	cpdef XmmsResult signal_playback_playtime(self, cb = None):
		"""
		Set a method to handle the playback playtime signal from the
		XMMS2 daemon.
		"""
		return self.create_result(cb, xmmsc_signal_playback_playtime(self.conn))

	cpdef XmmsResult playback_volume_set(self, channel, int volume, cb = None):
		"""
		Set the playback volume for specified channel
		"""
		c = from_unicode(channel)
		return self.create_result(cb, xmmsc_playback_volume_set(self.conn, c, volume))

	cpdef XmmsResult playback_volume_get(self, cb = None):
		"""
		Get the playback for all channels
		"""
		return self.create_result(cb, xmmsc_playback_volume_get(self.conn))

	cpdef XmmsResult broadcast_playback_volume_changed(self, cb = None):
		"""
		Set a broadcast callback for volume updates
		"""
		return self.create_result(cb, xmmsc_broadcast_playback_volume_changed(self.conn))

	cpdef XmmsResult broadcast_playlist_loaded(self, cb = None):
		"""
		Set a broadcast callback for loaded playlist event
		"""
		return self.create_result(cb, xmmsc_broadcast_playlist_loaded(self.conn))

	cpdef XmmsResult playlist_load(self, playlist, cb = None):
		"""
		Load the playlist as current playlist

		:return: The result of the operation.
		"""
		p = check_playlist(playlist, False)
		return self.create_result(cb, xmmsc_playlist_load(self.conn, <char *>p))

	cpdef XmmsResult playlist_list(self, cb = None):
		"""
		Lists the playlists

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_playlist_list(self.conn))

	cpdef XmmsResult playlist_remove(self, playlist, cb = None):
		"""
		Remove the playlist from the server

		:return: The result of the operation.
		"""
		p = check_playlist(playlist, False)
		return self.create_result(cb, xmmsc_playlist_remove(self.conn, <char *>p))

	cpdef XmmsResult playlist_shuffle(self, playlist = None, cb = None):
		"""
		Instruct the XMMS2 daemon to shuffle the playlist.

		:return: The result of the operation.
		"""
		p = check_playlist(playlist, True)
		return self.create_result(cb, xmmsc_playlist_shuffle(self.conn, <char *>p))

	cpdef XmmsResult playlist_rinsert(self, int pos, url, playlist = None, cb = None, encoded = False):
		"""
		Insert a directory in the playlist.
		Requires an int 'pos' and a string 'url' as argument.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res

		c = from_unicode(url)
		p = check_playlist(playlist, True)

		if encoded:
			res = xmmsc_playlist_rinsert_encoded(self.conn, <char *>p, pos, <char *>c)
		else:
			res = xmmsc_playlist_rinsert(self.conn, <char *>p, pos, <char *>c)
		return self.create_result(cb, res)

	@deprecated
	def playlist_rinsert_encoded(self, int pos, url, playlist = None, cb = None):
		"""
		:deprecated: Use `playlist_rinsert` instead.
		"""
		return self.playlist_rinsert(pos, url, playlist, cb = cb, encoded = True)

	cpdef XmmsResult playlist_insert_url(self, int pos, url, playlist = None, cb = None, encoded = False):
		"""
		Insert a path or URL to a playable media item to the playlist.
		Playable media items may be files or streams.
		Requires an int 'pos' and a string 'url' as argument.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res

		c = from_unicode(url)
		p = check_playlist(playlist, True)

		if encoded:
			res = xmmsc_playlist_insert_encoded(self.conn, <char *>p, pos, <char *>c)
		else:
			res = xmmsc_playlist_insert_url(self.conn, <char *>p, pos, <char *>c)
		return self.create_result(cb, res)

	@deprecated
	def playlist_insert_encoded(self, int pos, url, playlist = None, cb = None):
		"""
		:deprecated: Use `playlist_insert_url` instead.
		"""
		return self.playlist_insert_url(pos, url, playlist, cb = cb, encoded = True)


	cpdef XmmsResult playlist_insert_id(self, int pos, int id, playlist = None, cb = None):
		"""
		Insert a medialib to the playlist.
		Requires an int `pos` and an int `id` as argument.

		:return: The result of the operation.
		"""
		p = check_playlist(playlist, True)
		return self.create_result(cb, xmmsc_playlist_insert_id(self.conn, <char *>p, pos, id))


	cpdef XmmsResult playlist_insert_collection(self, int pos, Collection coll, order = None, playlist = None, cb = None):
		"""
		Insert the content of a collection to the playlist.
		Requires an int 'pos' and an int 'id' as argument.

		:return: The result of the operation.
		"""
		cdef xmmsv_t *order_val
		cdef xmmsc_result_t *res

		if order is None:
			order = []
		p = check_playlist(playlist, True)
		order_val = create_native_value(order)
		res = xmmsc_playlist_insert_collection(self.conn, <char *>p, pos, coll.coll, order_val)
		xmmsv_unref(order_val)
		return self.create_result(cb, res)

	cpdef XmmsResult playlist_radd(self, url, playlist = None, cb = None, encoded = False):
		"""
		Add a directory to the playlist.
		Requires a string 'url' as argument.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res

		c = from_unicode(url)
		p = check_playlist(playlist, True)
		if encoded:
			res = xmmsc_playlist_radd_encoded(self.conn, <char *>p, <char *>c)
		else:
			res = xmmsc_playlist_radd(self.conn, <char *>p, <char *>c)
		return self.create_result(cb, res)

	@deprecated
	def playlist_radd_encoded(self, url, playlist = None, cb = None):
		"""
		:deprecated: Use `playlist_radd` instead.
		"""
		return self.playlist_radd(url, playlist, cb = cb, encoded = True)

	cpdef XmmsResult playlist_add_url(self, url, playlist = None, cb = None, encoded = False):
		"""
		Add a path or URL to a playable media item to the playlist.
		Playable media items may be files or streams.
		Requires a string 'url' as argument.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res

		c = from_unicode(url)
		p = check_playlist(playlist, True)
		if encoded:
			res = xmmsc_playlist_add_encoded(self.conn, <char *>p, <char *>c)
		else:
			res = xmmsc_playlist_add_url(self.conn, <char *>p, <char *>c)
		return self.create_result(cb, res)

	@deprecated
	def playlist_add_encoded(self, url, playlist = None, cb = None):
		"""
		:deprecated: Use `playlist_add_url` instead.
		"""
		return self.playlist_add_url(url, playlist, cb = cb, encoded = True)

	cpdef XmmsResult playlist_add_id(self, int id, playlist = None, cb = None):
		"""
		Add a medialib id to the playlist.

		:return: The result of the operation.
		"""
		p = check_playlist(playlist, True)
		return self.create_result(cb, xmmsc_playlist_add_id(self.conn, <char *>p, id))

	cpdef XmmsResult playlist_add_collection(self, Collection coll, order = None, playlist = None, cb = None):
		"""
		Add the content of a collection to the playlist.

		:return: The result of the operation.
		"""
		cdef xmmsv_t *order_val
		cdef xmmsc_result_t *res

		p = check_playlist(playlist, True)
		if order is None:
			order = []
		order_val = create_native_value(order)
		res = xmmsc_playlist_add_collection(self.conn, <char *>p, coll.coll, order_val)
		xmmsv_unref(order_val)
		return self.create_result(cb, res)

	cpdef XmmsResult playlist_remove_entry(self, int id, playlist = None, cb = None):
		"""
		Remove a certain media item from the playlist.
		Requires a number 'id' as argument.

		:return: The result of the operation.
		"""
		p = check_playlist(playlist, True)
		return self.create_result(cb, xmmsc_playlist_remove_entry(self.conn, <char *>p, id))

	cpdef XmmsResult playlist_clear(self, playlist = None, cb = None):
		"""
		Clear the playlist.

		:return: The result of the operation.
		"""
		p = check_playlist(playlist, True)
		return self.create_result(cb, xmmsc_playlist_clear(self.conn, <char *>p))

	cpdef XmmsResult playlist_list_entries(self, playlist = None, cb = None):
		"""
		Get the current playlist. This function returns a list of IDs
		of the files/streams currently in the playlist. Use
		`medialib_get_info` to retrieve more specific information.

		:return: The current playlist.
		"""
		p = check_playlist(playlist, True)
		return self.create_result(cb, xmmsc_playlist_list_entries(self.conn, <char *>p))

	cpdef XmmsResult playlist_sort(self, props, playlist = None, cb = None):
		"""
		Sorts the playlist according to the properties specified.

		:return: The result of the operation.
		"""
		cdef xmmsv_t *props_val
		cdef xmmsc_result_t *res

		p = check_playlist(playlist, True)
		props_val = create_native_value(props)
		res = xmmsc_playlist_sort(self.conn, <char *>p, props_val)
		xmmsv_unref(props_val)
		return self.create_result(cb, res)

	cpdef XmmsResult playlist_set_next_rel(self, int position, cb = None):
		"""
		Sets the position in the playlist. Same as `playlist_set_next`
		but sets the next position relative to the current position.
		You can do set_next_rel(-1) to move backwards for example.
		"""
		return self.create_result(cb, xmmsc_playlist_set_next_rel(self.conn, position))

	cpdef XmmsResult playlist_set_next(self, int position, cb = None):
		"""
		Sets the position to move to, next, in the playlist. Calling
		`playback_tickle` will perform the jump to that position.
		"""
		return self.create_result(cb, xmmsc_playlist_set_next(self.conn, position))

	cpdef XmmsResult playlist_move(self, int cur_pos, int new_pos, playlist = None, cb = None):
		"""
		Moves a playlist entry to a new position.

		:return: The result of the operation.
		"""
		p = check_playlist(playlist, True)
		return self.create_result(cb, xmmsc_playlist_move_entry(self.conn, <char *>p, cur_pos, new_pos))

	cpdef XmmsResult playlist_create(self, playlist, cb = None):
		"""
		Create a new playlist.

		:return: The result of the operation.
		"""
		p = check_playlist(playlist, False)
		return self.create_result(cb, xmmsc_playlist_create(self.conn, <char *>p))

	cpdef XmmsResult playlist_current_pos(self, playlist = None, cb = None):
		"""
		Returns the current position in the playlist. This value will
		always be equal to, or larger than 0. The first entry in the
		list is 0.
		"""
		p = check_playlist(playlist, True)
		return self.create_result(cb, xmmsc_playlist_current_pos(self.conn, <char *>p))

	cpdef XmmsResult playlist_current_active(self, cb = None):
		"""
		Returns the name of the current active playlist
		"""
		return self.create_result(cb, xmmsc_playlist_current_active(self.conn))

	cpdef XmmsResult broadcast_playlist_current_pos(self, cb = None):
		"""
		Set a method to handle the playlist current position updates
		from the XMMS2 daemon. This is triggered whenever the daemon
		jumps from one playlist position to another. (not when moving
		a playlist item from one position to another)
		"""
		return self.create_result(cb, xmmsc_broadcast_playlist_current_pos(self.conn))

	cpdef XmmsResult broadcast_playlist_changed(self, cb = None):
		"""
		Set a method to handle the playlist changed broadcast from the
		XMMS2 daemon. Updated data is sent whenever the daemon's
		playlist changes.
		"""
		return self.create_result(cb, xmmsc_broadcast_playlist_changed(self.conn))

	cpdef XmmsResult broadcast_config_value_changed(self, cb = None):
		"""
		Set a method to handle the config value changed broadcast
		from the XMMS2 daemon.(i.e. some configuration value has
		been modified) Updated data is sent whenever a config
		value is modified.

		:return: The modified config key and its value.
		"""
		return self.create_result(cb, xmmsc_broadcast_config_value_changed(self.conn))

	cpdef XmmsResult config_set_value(self, key, val, cb = None):
		"""
		Set a configuration value on the daemon, given a key.

		:return: The result of the operation.
		"""
		k = from_unicode(key)
		v = from_unicode(val)
		return self.create_result(cb, xmmsc_config_set_value(self.conn, <char *>k, <char *>v))

	cpdef XmmsResult config_get_value(self, key, cb = None):
		"""
		Get the configuration value of a given key, from the daemon.

		:return: The result of the operation.
		"""
		k = from_unicode(key)
		return self.create_result(cb, xmmsc_config_get_value(self.conn, <char *>k))

	cpdef XmmsResult config_list_values(self, cb = None):
		"""
		Get list of configuration keys on the daemon. Use
		`config_get_value` to retrieve the values corresponding to the
		configuration keys.

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_config_list_values(self.conn))

	cpdef XmmsResult config_register_value(self, valuename, defaultvalue, cb = None):
		"""
		Register a new configvalue.
		This should be called in the initcode as XMMS2 won't allow
		set/get on values that haven't been registered.

		:return: The result of the operation.
		"""
		v = from_unicode(valuename)
		dv = from_unicode(defaultvalue)
		return self.create_result(cb, xmmsc_config_register_value(self.conn, <char *>v, <char *>dv))

	cpdef XmmsResult medialib_add_entry(self, path, cb = None, encoded = False):
		"""
		Add an entry to the MediaLib.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res
		p = from_unicode(path)
		if encoded:
			res = xmmsc_medialib_add_entry_encoded(self.conn, <char *>p)
		else:
			res = xmmsc_medialib_add_entry(self.conn, <char *>p)
		return self.create_result(cb, res)

	@deprecated
	def medialib_add_entry_encoded(self, path, cb = None):
		"""
		:deprecated: Use `medialib_add_entry` instead.
		"""
		return self.medialib_add_entry(path, cb = cb, encoded = True)

	cpdef XmmsResult medialib_remove_entry(self, int id, cb = None):
		"""
		Remove an entry from the medialib.

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_medialib_remove_entry(self.conn, id))

	cpdef XmmsResult medialib_move_entry(self, int id,  url, cb = None, encoded = False):
		"""
		Set a new url for an entry in the medialib.

		:return: The result of the operation.
		"""
		if encoded:
			try:
				from urllib import unquote_plus
			except ImportError: #Py3k
				from urllib.parse import unquote_plus
			url = unquote_plus(url)
		u = from_unicode(url)
		return self.create_result(cb, xmmsc_medialib_move_entry(self.conn, id, <char *>u))

	cpdef XmmsResult medialib_get_info(self, int id, cb = None):
		"""

		:return: Information about the medialib entry position specified.
		"""
		cdef XmmsResult res
		res = self.create_result(cb, xmmsc_medialib_get_info(self.conn, id))
		res.ispropdict = 1
		return res

	cpdef XmmsResult medialib_rehash(self, int id = 0, cb = None):
		"""
		Force the medialib to check that metadata stored is up to date.

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_medialib_rehash(self.conn, id))

	cpdef XmmsResult medialib_get_id(self, url, cb = None, encoded = False):
		"""
		Search for an entry (URL) in the medialib and return its ID number.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res
		u = from_unicode(url)
		if encoded:
			res = xmmsc_medialib_get_id_encoded(self.conn, <char *>u)
		else:
			res = xmmsc_medialib_get_id(self.conn, <char *>u)
		return self.create_result(cb, res)

	cpdef XmmsResult medialib_import_path(self, path, cb = None, encoded = False):
		"""
		Import metadata from all files recursively from the directory
		passed as argument.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res
		p = from_unicode(path)
		if encoded:
			res = xmmsc_medialib_import_path_encoded(self.conn, <char *>p)
		else:
			res = xmmsc_medialib_import_path(self.conn, <char *>p)
		return self.create_result(cb, res)

	@deprecated
	def medialib_path_import(self, path, cb = None, encoded = False):
		"""
		:deprecated: Use `medialib_import_path` instead.
		"""
		return self.medialib_import_path(path, cb = cb, encoded = encoded)

	@deprecated
	def medialib_path_import_encoded(self, path, cb = None):
		"""
		:deprecated: Use `medialib_import_path` instead.
		"""
		return self.medialib_import_path(path, cb = cb, encoded = True)

	cpdef XmmsResult medialib_property_set(self, int id, key, value, source = None, cb = None):
		"""
		Associate a value with a medialib entry. Source is optional.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res
		k = from_unicode(key)
		if isinstance(value, int):
			if source:
				s = from_unicode(source)
				res = xmmsc_medialib_entry_property_set_int_with_source(self.conn, id, <char *>s, <char *>k, value)
			else:
				res = xmmsc_medialib_entry_property_set_int(self.conn, id, <char *>k, value)
		else:
			v = from_unicode(value)
			if source:
				s = from_unicode(source)
				res = xmmsc_medialib_entry_property_set_str_with_source(self.conn, id, <char *>s, <char *>k, <char *>v)
			else:
				res = xmmsc_medialib_entry_property_set_str(self.conn, id, <char *>k, <char *>v)
		return self.create_result(cb, res)

	cpdef XmmsResult medialib_property_remove(self, int id, key, source = None, cb = None):
		"""
		Remove a value from a medialib entry. Source is optional.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res
		k = from_unicode(key)
		if source:
			s = from_unicode(source)
			res = xmmsc_medialib_entry_property_remove_with_source(self.conn, id, <char *>s, <char *>k)
		else:
			res = xmmsc_medialib_entry_property_remove(self.conn, id, <char *>k)
		return self.create_result(cb, res)

	cpdef XmmsResult broadcast_medialib_entry_added(self, cb = None):
		"""
		Set a method to handle the medialib entry added broadcast
		from the XMMS2 daemon. (i.e. a new entry has been added)
		"""
		return self.create_result(cb, xmmsc_broadcast_medialib_entry_added(self.conn))

	@deprecated
	def broadcast_medialib_entry_changed(self, cb = None):
		"""
		:deprecated: Use `broadcast_medialib_entry_updated` instead.
		"""
		return self.broadcast_medialib_entry_updated(cb)

	cpdef XmmsResult broadcast_medialib_entry_updated(self, cb = None):
		"""
		Set a method to handle the medialib entry updated broadcast
		from the XMMS2 daemon.
		Updated data is sent when the metadata for a song is updated
		in the medialib.
		"""
		return self.create_result(cb, xmmsc_broadcast_medialib_entry_updated(self.conn))

	cpdef XmmsResult broadcast_medialib_entry_removed(self, cb = None):
		"""
		Set a method to handle the medialib entry removed broadcast
		from the XMMS2 daemon. (i.e. an entry has been removed)
		"""
		return self.create_result(cb, xmmsc_broadcast_medialib_entry_removed(self.conn))

	cpdef XmmsResult broadcast_collection_changed(self, cb = None):
		"""
		Set a method to handle the collection changed broadcast
		from the XMMS2 daemon.
		"""
		return self.create_result(cb, xmmsc_broadcast_collection_changed(self.conn))

	cpdef XmmsResult signal_mediainfo_reader_unindexed(self, cb = None):
		"""
		Tell daemon to send you the number of unindexed files in the mlib

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_signal_mediainfo_reader_unindexed(self.conn))

	cpdef XmmsResult broadcast_mediainfo_reader_status(self, cb = None):
		"""
		Tell daemon to send you the status of the mediainfo reader

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_broadcast_mediainfo_reader_status(self.conn))

	cpdef XmmsResult xform_media_browse(self, url, cb = None, encoded = False):
		"""
		Browse files from xform plugins.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res
		u = from_unicode(url)
		if encoded:
			res = xmmsc_xform_media_browse_encoded(self.conn, <char *>u)
		else:
			res = xmmsc_xform_media_browse(self.conn, <char *>u)
		return self.create_result(cb, res)

	@deprecated
	def xform_media_browse_encoded(self, url, cb = None):
		"""
		:deprecated: Use `xform_media_browse` instead.
		"""
		return self.xform_media_browse(url, cb = cb, encoded = True)

	cpdef XmmsResult coll_get(self, name, ns = "Collections", cb = None):
		"""
		Retrieve a Collection

		:return: The result of the operation.
		"""
		cdef char *n
		n = check_namespace(ns, False)
		nam = from_unicode(name)
		return self.create_result(cb, xmmsc_coll_get(self.conn, nam, n))

	cpdef XmmsResult coll_list(self, ns = "Collections", cb = None):
		"""
		List collections

		:return: The result of the operation.
		"""
		cdef char *n
		n = check_namespace(ns, False)
		return self.create_result(cb, xmmsc_coll_list(self.conn, n))

	cpdef XmmsResult coll_save(self, Collection coll, name, ns = "Collections", cb = None):
		"""
		Save a collection on server.

		:return: The result of the operation.
		"""
		cdef char *n
		n = check_namespace(ns, False)
		nam = from_unicode(name)
		return self.create_result(cb, xmmsc_coll_save(self.conn, coll.coll, <char *>nam, n))

	cpdef XmmsResult coll_remove(self, name, ns = "Collections", cb = None):
		"""
		Remove a collection on server.

		:return: The result of the operation.
		"""
		cdef char *n
		n = check_namespace(ns, False)
		nam = from_unicode(name)
		return self.create_result(cb, xmmsc_coll_remove(self.conn, <char *>nam, n))

	cpdef XmmsResult coll_rename(self, oldname, newname, ns = "Collections", cb = None):
		"""
		Rename a collection.

		:return: The result of the operation.
		"""
		cdef char *n
		n = check_namespace(ns, False)

		oldnam = from_unicode(oldname)
		newnam = from_unicode(newname)
		return self.create_result(cb, xmmsc_coll_rename(self.conn, <char *>oldnam, <char *>newnam, n))

	cpdef XmmsResult coll_idlist_from_playlist_file(self, path, cb = None):
		"""
		Create an idlist from a playlist.

		:return: The result of the operation.
		"""
		p = from_unicode(path)
		return self.create_result(cb, xmmsc_coll_idlist_from_playlist_file(self.conn, <char *>p))

	cpdef XmmsResult coll_query(self, Collection coll, fetch, cb = None):
		"""
		Retrive a list of ids of the media matching the collection

		:return: The result of the operation.
		"""
		cdef xmmsv_t *fetch_val
		fetch_val = create_native_value(fetch)
		res = self.create_result(cb, xmmsc_coll_query(self.conn, coll.coll, fetch_val))
		return res

	cpdef XmmsResult coll_query_ids(self, Collection coll, start = 0, leng = 0, order = None, cb = None):
		"""
		Retrive a list of ids of the media matching the collection

		:return: The result of the operation.
		"""
		cdef xmmsv_t *order_val
		cdef xmmsc_result_t *res

		if order is None:
			order = []
		order_val = create_native_value(order)
		res = xmmsc_coll_query_ids(self.conn, coll.coll, order_val, start, leng)
		xmmsv_unref(order_val)
		return self.create_result(cb, res)

	cpdef XmmsResult coll_query_infos(self, Collection coll, fields, start = 0, leng = 0, order = None, groupby = None, cb = None):
		"""
		Retrive a list of mediainfo of the media matching the collection

		:return: The result of the operation.
		"""
		cdef xmmsv_t *order_val
		cdef xmmsv_t *fields_val
		cdef xmmsv_t *groupby_val
		cdef xmmsc_result_t *res

		if order is None:
			order = []
		if groupby is None:
			groupby = []
		order_val = create_native_value(order)
		fields_val = create_native_value(fields)
		groupby_val = create_native_value(groupby)
		res = xmmsc_coll_query_infos(self.conn, coll.coll, order_val, start, leng, fields_val, groupby_val)
		xmmsv_unref(order_val)
		xmmsv_unref(fields_val)
		xmmsv_unref(groupby_val)
		return self.create_result(cb, res)

	cpdef XmmsResult c2c_ready(self, cb = None):
		"""
		Notify the server that client services are ready for query.
		This method is called `XmmsServiceNamespace.register()` and don't need to
		be called explicitly.

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res

		if not cb:
			cb = _noop

		res = xmmsc_c2c_ready(self.conn)
		return self.create_result(cb, res)

	cpdef XmmsResult c2c_get_connected_clients(self, cb = None):
		"""
		Get a list of clients connected to the xmms2 server

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res

		res = xmmsc_c2c_get_connected_clients (self.conn)
		return self.create_result(cb, res)

	cpdef XmmsResult c2c_get_ready_clients(self, cb = None):
		"""
		Get a list of clients connected to the xmms2 server

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res

		res = xmmsc_c2c_get_ready_clients (self.conn)
		return self.create_result(cb, res)

	cpdef XmmsResult broadcast_c2c_ready(self, cb = None):
		"""
		Broadcast reveiced whenever a client's service api is ready

		:return: the result of the operation.
		"""
		cdef xmmsc_result_t *res

		res = xmmsc_broadcast_c2c_ready(self.conn)
		return self.create_result(cb, res)


	cpdef XmmsResult broadcast_c2c_client_connected(self, cb = None):
		"""
		Broadcast received whenever a new client connects to the server

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res

		res = xmmsc_broadcast_c2c_client_connected(self.conn)
		return self.create_result(cb, res)

	cpdef XmmsResult broadcast_c2c_client_disconnected(self, cb = None):
		"""
		Broadcast received whenever a client disconnects from the server

		:return: The result of the operation.
		"""
		cdef xmmsc_result_t *res

		res = xmmsc_broadcast_c2c_client_disconnected(self.conn)
		return self.create_result(cb, res)

	cpdef bint sc_init(self):
		"""
		Initialize client-to-client features.
		Client-to-client features won't work with the synchronous client wrapper

		:return: whether client-to-client is initialized.
		"""
		return xmmsc_sc_init(self.conn) != NULL

	cpdef bint sc_broadcast_emit(self, broadcast, value = None):
		"""
		Emit a broadcast message to subscribed clients

		:return: whether notifications were successfully queued
		"""
		cdef xmmsv_t *val
		cdef xmmsv_t *bc_path
		cdef bint ret

		bc_path = create_native_value(broadcast)
		val = create_native_value(value)
		ret = xmmsc_sc_broadcast_emit(self.conn, bc_path, val)
		xmmsv_unref(bc_path)
		xmmsv_unref(val)
		return ret

	cpdef XmmsResult sc_broadcast_subscribe(self, int dest, broadcast, cb = None):
		"""
		Subscribe to a broadcast from another client

		:return: The result of the operation.
		"""
		cdef xmmsv_t *bc_path
		cdef xmmsc_result_t *res

		bc_path = create_native_value(broadcast)
		res = xmmsc_sc_broadcast_subscribe(self.conn, dest, bc_path)
		xmmsv_unref(bc_path)
		return self.create_result(cb, res)

	cpdef XmmsResult sc_call(self, int dest, method, args = (), kargs = dict(), cb = None):
		"""
		Call a remote method

		:return: The result of the operation.
		"""
		cdef xmmsv_t *m_path
		cdef xmmsv_t *m_pos
		cdef xmmsv_t *m_named
		cdef xmmsc_result_t *res

		m_path = create_native_value(method)
		m_pos = create_native_value(args)
		m_named = create_native_value(kargs)
		res = xmmsc_sc_call(self.conn, dest, m_path, m_pos, m_named)
		xmmsv_unref(m_path)
		xmmsv_unref(m_pos)
		xmmsv_unref(m_named)
		return self.create_result(cb, res)

	cpdef XmmsResult sc_introspect_namespace(self, int dest, path = (), cb = None):
		"""
		Get informations about a namespace on a remote client

		:return: The result of the operation.
		"""
		cdef xmmsv_t *i_path
		cdef xmmsc_result_t *res

		i_path = create_native_value(path)
		res = xmmsc_sc_introspect_namespace(self.conn, dest, i_path)
		xmmsv_unref(i_path)
		return self.create_result(cb, res)

	cpdef XmmsResult sc_introspect_method(self, int dest, path, cb = None):
		"""
		Get informations about a method on a remote client

		:return: The result of the operation.
		"""
		cdef xmmsv_t *i_path
		cdef xmmsc_result_t *res

		i_path = create_native_value(path)
		res = xmmsc_sc_introspect_method(self.conn, dest, i_path)
		xmmsv_unref(i_path)
		return self.create_result(cb, res)

	cpdef XmmsResult sc_introspect_broadcast(self, int dest, path, cb = None):
		"""
		Get informations about a broadcast on a remote client

		:return: The result of the operation.
		"""
		cdef xmmsv_t *i_path
		cdef xmmsc_result_t *res

		i_path = create_native_value(path)
		res = xmmsc_sc_introspect_broadcast(self.conn, dest, i_path)
		xmmsv_unref(i_path)
		return self.create_result(cb, res)

	cpdef XmmsResult sc_introspect_constant(self, int dest, path, cb = None):
		"""
		Get informations about a constant on a remote client

		:return: The result of the operation.
		"""
		cdef xmmsv_t *i_path
		cdef xmmsc_result_t *res
		ns = path[:-1]
		key = from_unicode(path[-1])
		i_path = create_native_value(ns)
		res = xmmsc_sc_introspect_constant(self.conn, dest, i_path, <char *>key)
		xmmsv_unref(i_path)
		return self.create_result(cb, res)

	cpdef XmmsResult sc_introspect_docstring(self, int dest, path, cb = None):
		"""
		Get the docstring for a given path on a remote client

		:return: The result of the operation.
		"""
		cdef xmmsv_t *i_path
		cdef xmmsc_result_t *res

		i_path = create_native_value(path)
		res = xmmsc_sc_introspect_docstring(self.conn, dest, i_path)
		xmmsv_unref(i_path)
		return self.create_result(cb, res)

	cpdef XmmsResult bindata_add(self, data, cb = None):
		"""
		Add a datafile to the server

		:return: The result of the operation.
		"""
		cdef char *t
		t = <char *>data
		return self.create_result(cb, xmmsc_bindata_add(self.conn,<unsigned char *>t,len(data)))

	cpdef XmmsResult bindata_retrieve(self, hash, cb = None):
		"""
		Retrieve a datafile from the server

		:return: The result of the operation.
		"""
		h = from_unicode(hash)
		return self.create_result(cb, xmmsc_bindata_retrieve(self.conn, <char *>h))

	cpdef XmmsResult bindata_remove(self, hash, cb = None):
		"""
		Remove a datafile from the server

		:return: The result of the operation.
		"""
		h = from_unicode(hash)
		return self.create_result(cb, xmmsc_bindata_remove(self.conn, <char *>h))

	cpdef XmmsResult bindata_list(self, cb = None):
		"""
		List all bindata hashes stored on the server

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_bindata_list(self.conn))

	cpdef XmmsResult stats(self, cb = None):
		"""
		Get statistics information from the server

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_main_stats(self.conn))

	cpdef XmmsResult visualization_version(self, cb = None):
		"""
		Get the version of the visualization plugin installed on the server.

		:return: The result of the operation.
		"""
		return self.create_result(cb, xmmsc_visualization_version(self.conn))

	cpdef XmmsResult visualization_init(self, cb = None):
		"""
		Get a new visualization handle.

		:return: The result of the operation
		"""
		return self.create_vis_result(cb, xmmsc_visualization_init(self.conn), VIS_RESULT_CMD_INIT)

	cpdef XmmsResult visualization_start(self, int handle, cb = None):
		"""
		Starts the visualization.

		:return: The result of the operation
		"""
		return self.create_vis_result(cb, xmmsc_visualization_start(self.conn, handle), VIS_RESULT_CMD_START)

	cpdef bint visualization_started(self, int handle):
		"""
		Whether the visualization is started or not.

		:return: True if the visualization is started, False otherwise.
		"""
		return xmmsc_visualization_started(self.conn, handle)

	cpdef bint visualization_errored(self, int handle):
		"""
		Whether the visualization got an error.

		:return: True if the visualization got an error, False otherwise.
		"""
		return xmmsc_visualization_errored(self.conn, handle)

	cpdef XmmsResult visualization_property_set(self, int handle, key, value, cb = None):
		"""
		Set a visualization's property.

		:return: The result of the operation
		"""
		k = from_unicode(key)
		v = from_unicode(value)
		return self.create_result(cb, xmmsc_visualization_property_set(self.conn, handle, <char *>k, <char *>v))

	cpdef XmmsResult visualization_properties_set(self, int handle, props = {}, cb = None):
		"""
		Set visualization's properties.

		:return: The result of the operation
		"""
		cdef xmmsv_t *_props
		cdef xmmsc_result_t *res

		_props = create_native_value(props)
		res = xmmsc_visualization_properties_set(self.conn, handle, _props)
		xmmsv_unref(_props)
		return self.create_result(cb, res)

	cpdef XmmsVisChunk visualization_chunk_get(self, int handle, int drawtime = 0, bint blocking = False):
		"""
		Fetches the next available data chunk


		:return: Visualization chunk.
		"""
		cdef short *buf
		cdef int size
		cdef XmmsVisChunk chunk

		buf = <short *>PyMem_Malloc(2 * XMMSC_VISUALIZATION_WINDOW_SIZE * sizeof (short))
		size = xmmsc_visualization_chunk_get (self.conn, handle, buf, drawtime, blocking)
		if size < 0:
			PyMem_Free(buf)
			raise VisualizationError("Unrecoverable error in visualization")
		chunk = XmmsVisChunk()
		chunk.set_data(buf, size)
		PyMem_Free(buf)
		return chunk

	cpdef visualization_shutdown(self, int handle):
		"""
		Shutdown and destroy a visualization. After this, handle is no longer valid.
		"""
		xmmsc_visualization_shutdown(self.conn, handle)



class XmmsDisconnectException(Exception):
	pass

cdef class XmmsLoop(XmmsApi):
	"""
	A simple main loop to drive async communication.

	>>> import xmmsclient
	>>> def handle_stats(result):
	>>>   print(result.value())
	>>> xc = xmmsclient.XmmsLoop('clientname')
	>>> xc.connect()
	>>> xc.stats(cb=handle_stats)
	>>> xc.loop()
	"""
	#cdef bint do_loop
	#cdef object wakeup

	def __cinit__(self):
		self.do_loop = 0

	def _loop_set_wakeup(self, fd):
		"""
		Set the file descriptor for wakeup operations.
		This method should not be used outside the XmmsLoop class implementation.
		It is provided to allow subclasses to add some locking on it in
		multithreaded applications.

		Use a Rlock if you want to synchronize other methods using the same lock.
		"""
		self.wakeup = fd
	def _loop_get_wakeup(self):
		"""
		Get the file descriptor for wakeup operations.
		This method should not be used outside the XmmsLoop class implementation.
		It is provided to allow subclasses to add some locking on it in
		multithreaded applications.

		Use a Rlock if you want to synchronize other methods using the same lock.
		"""
		return self.wakeup

	def exit_loop(self):
		"""
		Exits from the `loop` call
		"""
		self.do_loop = False
		self.loop_tickle()

	def loop_tickle(self):
		from os import write
		w = self._loop_get_wakeup()
		if w is not None:
			write(w, "1".encode('ascii'))

	def loop_iter(self, infd = None, outfd = None, errfd = None, timeout = -1):
		"""
		Run one iteration of the main loop. Should be overridden to add
		custom operations in the main loop.

		:return: The tuple returned by `select.select` to be used by overridding
		         methods in subclasses.
		"""
		cdef int fd
		fd = xmmsc_io_fd_get(self.conn)
		if self.want_ioout():
			w = [fd]
		else:
			w = []
		r = [fd]
		err = [fd]
		if infd:
			r += infd
		if outfd:
			w += outfd
		if errfd:
			err += errfd
		if timeout < 0:
			(i, o, e) = select(r, w, err)
		else:
			(i, o, e) = select(r, w, err, timeout)
		if e and fd in e:
			xmmsc_io_disconnect(self.conn)
			raise XmmsDisconnectException()
		if i and fd in i:
			if not xmmsc_io_in_handle(self.conn):
				raise XmmsDisconnectException()
		if o and fd in o:
			if not xmmsc_io_out_handle(self.conn):
				raise XmmsDisconnectException()
		return (i, o, e) #Can be used by overridding methods for extra handling.

	def loop(self):
		"""
		Main client loop for most python clients. Call this to run the
		client once everything has been set up. This function blocks
		until `exit_loop` is called. One can override `loop_iter` to
		perform extra processing.
		"""
		from os import pipe, read
		(r, w) = pipe()

		self.do_loop = True
		self._loop_set_wakeup(w)

		while self.do_loop:
			try:
				(i, o, e) = self.loop_iter(infd = [r])
				if r in i:
					read(r, 1) # Purge wakeup stream (each wakeup signal should not write more than one byte)
			except XmmsDisconnectException:
				self.do_loop = False

		self._loop_set_wakeup(None)

Xmms = XmmsApi



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess
import sys

if len(sys.argv) != 2 and len(sys.argv) != 3:
    print("usage: %s <from> [<to>]" % sys.argv[0])
    raise SystemExit(1)

rev_start = sys.argv[1]

if len(sys.argv) > 2:
    rev_end = sys.argv[2]
else:
    rev_end = "HEAD"

command = ["git", "log", "--pretty=%an", rev_start + "..." + rev_end]

authors = {}
for line in subprocess.check_output(command).decode('utf-8').split("\n"):
    if not line.strip():
        continue
    authors[line.strip()] = True

print(", ".join(sorted(authors.keys())))



#! /usr/bin/env python
# encoding: utf-8

import os
import re
from waflib import Task,Errors,Utils
from waflib.Configure import conf
from waflib.TaskGen import extension,feature,before_method


@feature('perldoc')
@before_method('process_source')
def perldoc_file(self):
	source = self.to_nodes(getattr(self, 'source', []))
	self.source = []

	for node in source:
		if not node:
			raise Errors.BuildError('cannot find input file for processing')

		match = re.search(r"MODULE\s*=\s*(\S+)", node.read())
		if match is None:
			raise Errors.BuildError('.xs file does not contain a MODULE declaration')
		module = match.group(1)

		path = module.replace("::", os.sep) + ".pod"
		outnode = node.parent.find_or_declare(path)
		self.create_task('perldoc',node,outnode)

		base_path = getattr(self, 'install_path', "${INSTALLDIR_PERL_LIB}")
		self.bld.install_files(os.path.join(base_path, os.path.dirname(path)), outnode)

class perldoc(Task.Task):
	run_str='${PODSELECT} ${SRC} > ${TGT}'
	color='BLUE'
	ext_in=['.xs']
	ext_out=['.pod']

def configure(ctx):
	ctx.find_program('podselect', var='PODSELECT')



#needs to reimport required symbols from .h files.

cdef extern from "xmmsc/xmmsv.h":
	ctypedef struct xmmsv_t
	ctypedef struct xmmsv_dict_iter_t
	ctypedef struct xmmsv_list_iter_t

cdef class XmmsValue:
	cdef object sourcepref
	cdef xmmsv_t *val
	cdef int ispropdict

	cdef  set_value(self, xmmsv_t *value, int ispropdict=*)
	cpdef set_pyval(self, pyval)
	cpdef get_type(self)
	cpdef iserror(self)
	cpdef is_error(self)
	cpdef get_error(self)
	cpdef get_int(self)
	cpdef get_float(self)
	cpdef get_string(self)
	cpdef get_bin(self)
	cpdef get_coll(self)
	cpdef get_dict(self)
	cpdef get_dict_iter(self)
	cpdef get_propdict(self)
	cpdef get_list(self)
	cpdef get_list_iter(self)
	cpdef value(self)
	cpdef copy(self, cls=*)

cdef class XmmsValueC2C(XmmsValue):
	pass

cdef class XmmsListIter:
	cdef object sourcepref
	cdef xmmsv_t *val
	cdef xmmsv_list_iter_t *it

cdef class XmmsDictIter:
	cdef object sourcepref
	cdef xmmsv_t *val
	cdef xmmsv_dict_iter_t *it

cdef class CollectionRef:
	cdef xmmsv_t *coll

	cdef set_collection(self, xmmsv_t *coll)

cdef class Collection(CollectionRef):
	cdef object _attributes
	cdef object _operands
	cdef object _idlist

	cdef init_idlist(self)
	cdef init_attributes(self)
	cdef init_operands(self)

	cpdef copy(self)

cdef class CollectionAttributes(CollectionRef):
	cpdef get_dict(self)
	cpdef iterkeys(self)
	cpdef keys(self)
	cpdef itervalues(self)
	cpdef iterxvalues(self)
	cpdef values(self)
	cpdef xvalues(self)
	cpdef iteritems(self)
	cpdef iterxitems(self)
	cpdef items(self)
	cpdef xitems(self)
	cpdef clear(self)

cdef enum AttributesIterType:
	ITER_KEYS = 1
	ITER_VALUES = 2
	ITER_ITEMS = 3
	ITER_XVALUES = 4
	ITER_XITEMS = 5
cdef class AttributesIterator:
	cdef xmmsv_dict_iter_t *diter
	cdef int itertype
	cdef CollectionAttributes attr

	cpdef reset(self)


cdef class CollectionOperands(CollectionRef):
	cdef object pylist
	cdef init_pylist(self)

	cpdef append(self, Collection op)
	cpdef extend(self, ops)
	cpdef remove(self, Collection op)
	cpdef clear(self)
	cpdef list(self)


cdef class CollectionIDList(CollectionRef):
	cpdef append(self, int v)
	cpdef extend(self, v)
	cpdef insert(self, int v, int i)
	cpdef remove(self, int i)
	cpdef pop(self, int i = *)
	cpdef clear(self)

cdef create_coll(xmmsv_t *coll)
cdef xmmsv_t *create_native_value(value) except NULL
cdef get_default_source_pref()



#!/usr/bin/env python
#  XMMS2 - X Music Multiplexer System
#  Copyright (C) 2003-2006 XMMS2 Team
# 
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#                   
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  This file is a part of the XMMS2 client tutorial #1

# include xmmsclient module
import xmmsclient
import os
import sys

"""
To communicate with xmms2d, you need an instance of the
xmmsclient.XMMS object, which abstracts the connection.
First you need to initialize the connection; as argument
you need to pass "name" of your client. The name has to
be in the range [a-zA-Z0-9] because xmms is deriving
configuration values from this name.
"""
xmms = xmmsclient.XMMS("tutorial1")

"""
Now we need to connect to xmms2d. We need to
pass the XMMS ipc-path to the connect call.
If passed None, it will default to 
unix:///tmp/xmms-ipc-<user>, but all xmms2 clients
should handle the XMMS_PATH enviroment in
order to configure connection path.

XMMS.connect will throw IOError if an error occured.
"""
try:
	xmms.connect(os.getenv("XMMS_PATH"))
except IOError, detail:
	print "Connection failed:", detail
	sys.exit(1)

"""
This is all you have to do to connect to xmms2d.
Now we can send commands. Let's do something easy
like getting xmms2d to start playback.

xmmsclient.XMMSResult is the type of object returned
from all commands that are given to the xmms2d server.
"""
result = xmms.playback_start()

"""
The command will be sent, and since this is a
synchronous connection we can block for its 
return here. The async / sync issue will be
commented on later.
"""
result.wait()

"""
When XMMSResult.wait() returns, we have the
answer from the server. Let's check for errors
and print it out if something went wrong
"""
if result.iserror():
	print "playback start returned error, %s" % result.get_error()
else:
	"""
	Now we are done, We can simply exit - Python's garbage
	collector will take care of the cleanup.
	(Alternatively, you can delete the XMMS instance
	using del)
	"""
	sys.exit(0)



from waflib import Task, Logs, Errors, Options, Utils
from waflib.TaskGen import feature, before_method
import os
import re

class create_test_runner(Task.Task):
    def generate_runner_source_code(self):
        TEST_CASE_PREFIX = "__testcase_wrapper_"

        declare = []
        register = []
        for i, (suite, tests) in enumerate(self.suites):
            register.append("""CU_pSuite suite%d = CU_add_suite ("%s", NULL, NULL);""" % (i, suite))
            for test in sorted(tests):
                register.append("""CU_add_test (suite%d, "%s", %s);""" % (i, test, TEST_CASE_PREFIX + test))
                declare.append("""void %s (void);""" % (TEST_CASE_PREFIX + test))

        return "\n\t".join(register), "\n".join(declare)

    def run(self):
        register, declare = self.generate_runner_source_code()
        code = self.inputs[0].read()
        code = code.replace("@@DECLARE_TEST_CASES@@", declare)
        code = code.replace("@@REGISTER_TEST_SUITES@@", register)
        self.outputs[0].write(code)

scraper = re.compile(r'^(CASE|SETUP)\s*\(([^)]+)\)')

def scrape_test_cases(node):
    suite = ""
    tests = []
    for line in node.read().split("\n"):
        match = scraper.match(line)
        if not match:
            continue
        typ, name = match.groups()
        if typ == "SETUP":
            suite = name
        elif typ == "CASE":
            tests.append(name)
    if suite and tests:
        return suite, tests
    return False

@feature("test")
@before_method("process_source")
def generate_runner(self):
    self.source = self.to_nodes(self.source)
    try:
        self.use = self.to_list(self.use)
    except (AttributeError):
        self.use = []

    suites = []
    for node in self.source:
        result = scrape_test_cases(node)
        if result:
            suites.append(result)

    if suites:
        # Generate and add runner
        test_runner_template = os.path.join("tests", "runner", "main.c")
        runner = self.bld.srcnode.find_resource(test_runner_template)
        target = self.path.find_or_declare("test_runner_%s.c" % self.target)

        task = self.create_task("create_test_runner", [runner] + self.source, target)
        task.suites = suites

        self.source += [target]

        # link valgrind_object in
        self.use += ["memorystatus"]


def monkey_patch_test_runner():
    original = Task.classes["utest"].run

    def xmms_test_runner(self):
        if self.env.enable_gcov:
            # If profiling is enabled, we don't want to invoke valgrind.
            return original(self)

        if getattr(self.generator, 'use_valgrind', self.env.VALGRIND and True):
            # Disable GLib memory optimizations to avoid Valgrind confusion.
            os.environ["G_SLICE"] = "always-malloc"
            os.environ["G_DEBUG"] = "gc-friendly"

            suppression = os.path.join(os.getcwd(), "utils", "valgrind-suppressions")
            self.generator.ut_exec = [
                self.env.VALGRIND[0],
                "--log-file=%s.log" % self.inputs[0].abspath(),
                "--leak-check=full",
                "--suppressions=%s" % suppression,
                "--num-callers=20",
                "--malloc-fill=0xa1",
                "--free-fill=0xa1",
                self.inputs[0].abspath()
            ]
        original(self)

    Task.classes["utest"].run = xmms_test_runner

def summary(bld):
    lst = getattr(bld, 'utest_results', [])
    if lst:
        Logs.pprint('CYAN', 'test summary')

        failed_tests = []

        for (f, code, out, err) in lst:
            if code != 0:
                failed_tests.append(f)
            Logs.pprint('NORMAL', out.decode("ascii", "ignore"))

        if len(failed_tests) > 0:
            raise Errors.WafError("Test(s) failed:\n%s" % "\n".join(failed_tests))

def run(cmd, cwd):
    proc = Utils.subprocess.Popen(cmd, cwd=cwd,
                                  stderr=Utils.subprocess.PIPE,
                                  stdout=Utils.subprocess.PIPE)
    (stdout, stderr) = proc.communicate()
    return (stdout, stderr, proc.returncode)

def generate_coverage(bld):
    if Options.options.generate_coverage:
        if not bld.env.enable_gcov:
            raise Errors.WafError("Coverage reports need --enable-gcov passed to configure.")

        if not (bld.env.LCOV and bld.env.GENHTML):
            raise Errors.WafError("Could not generate coverage as the tools are missing.")

        cmd = [bld.env.LCOV[0], "-c", "-b", ".", "-d", ".", "-o", "coverage.info"]
        (stdout, stderr, code) = run(cmd, bld.bldnode.abspath())

        if code != 0:
            err = "stdout: %s\nstderr: %s" % (stdout, stderr)
            raise Errors.WafError("Could not run coverage analysis tool!\n%s" % err)

        cmd = [bld.env.GENHTML[0], "-o", "coverage", "coverage.info"]
        (stdout, stderr, code) = run(cmd, bld.bldnode.abspath())

        if code != 0:
            err = "stdout: %s\nstderr: %s" % (stdout, stderr)
            raise Errors.WafError("Could not run coverage report tool!\n%s" % err)

def configure(conf):
    conf.load("waf_unit_test")
    if conf.options.enable_valgrind is True:
        conf.find_program("valgrind", var="VALGRIND", mandatory=False)
    conf.find_program("lcov", var="LCOV", mandatory=False)
    conf.find_program("genhtml", var="GENHTML", mandatory=False)

def setup(bld):
    monkey_patch_test_runner()
    bld.add_post_fun(generate_coverage)
    bld.add_post_fun(summary)

def options(opts):
    opts.load("waf_unit_test")
    opts.add_option('--generate-coverage', action='store_true',
                    dest='generate_coverage', default=False,
                    help="Generate coverage report.")



#!/usr/bin/env python
#  XMMS2 - X Music Multiplexer System
#  Copyright (C) 2003-2006 XMMS2 Team
# 
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#                   
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  This file is a part of the XMMS2 client tutorial #5
#  More about dicts and propdicts

import xmmsclient
import os
import sys

"""
The first part of this program is
commented in tut1.
"""
xmms = xmmsclient.XMMS("tutorial5")
try:
	xmms.connect(os.getenv("XMMS_PATH"))
except IOError, detail:
	print "Connection failed:", detail
	sys.exit(1)

"""
In tut3 we learned about dicts. But there is more to know on this
topic. There are actually two kinds of dicts. The normal ones and 
property dicts. I will try to explain them here.

A normal dict contains key:value mappings as normal. Getting values from
this is straight forward: just access them as normal python dicts as
we did in tut3.

Property dicts are dicts that can have the same key multiple times.
Like two "artists" or "titles". If a propdict is treated like a dict,
it will return only one of the possible values for a key. For example,
mypropdict["artist"] will return only one possible value for 'artist'
To get the values for keys from a specific source, use
PropDict.set_source_preference() Alternatively, you can access a
value associated to a key from a specific source by using a tuple
as the key: mypropdict[("server", "artist")] The tuple key always
has the source as first item and the real key name as second item.

Property dicts are primarily used by the medialib. In this case the
source refers to the application which set the tag.

Most of the time you don't have to care because the default source is
set to prefer values set by the server over values set by clients. But
if your program wants to override title or artist, for example, you need
to call PropDict.set_source_preference() before extracting values.

It's also important when iterating over the dicts. Let me show you.

First we retrieve the config values stored in the server and print
them out. This is a normal dict.
"""
result = xmms.configval_list()
result.wait()

for key, value in result.value().items():
	print key, "=", value

"""
Now get a prop dict. Entry 1 should be the default clip
we ship so it should be safe to request information
about it.
"""
result = xmms.medialib_get_info(1)
result.wait()
for key, value in result.value().items():
	print "%s:%s" % key, "=", value



# encoding: utf-8

import sys
import genipc
from indenter import Indenter

HEADER_PROLOG = """
/*  XMMS2 - X Music Multiplexer System
 *  Copyright (C) 2003-2023 XMMS2 Team
 *
 *  PLUGINS ARE NOT CONSIDERED TO BE DERIVED WORK !!!
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 */

/* Header generated from src/ipc.xml - DO NOT EDIT ! */

#ifndef __IPC_SIGNAL_XMMS_H__
#define __IPC_SIGNAL_XMMS_H__

#include <xmmsc/xmmsc_compiler.h>
""".strip()

HEADER_EPILOG = """
#include <xmmsc/xmmsc_idnumbers_extra.h>

#endif /* __IPC_SIGNAL_XMMS_H__ */
""".strip()

def _constant_name(name):
	return 'XMMS_%s' % name.upper()

def _c_type(name):
	return 'xmms_%s_t' % name

_escape = '\a\b\t\n\v\f\r"\\'
_escape_c = 'abtnvfr"\\'
def _c_string(s):
	ret = []
	for c in s:
		oc = ord(c)
		n = _escape.find(c)
		if n > -1:
			c = '\\%c' % _escape_c[n]
		elif oc < 32 or 0x7f <= oc <= 0xff:
			c = '\\x%02x' % oc
		elif oc > 0xff:
			raise ValueError("Unicode characters not supported.")
		ret.append(c)
	return '"%s"' % "".join(ret)

def _c_literal(val):
	""" Encode simple values in their C literal representation.
	Support int, float and str only.

	warning: Strings should contain only characters with ordinals <= 255.
	"""
	if isinstance(val, int):
		return "%d%s" % (val, (val >= 2**32 or val < -2**31) and "L" or "")
	elif isinstance(val, float):
		return ("%0.12f%s" % val).rstrip('0')
	elif isinstance(val, str):
		return _c_string(val)
	else:
		raise ValueError("Can't convert %r to C literal." % type(val).__name__)

def build_enum(enum, enums):
	if not enum:
		return
	Indenter.enter('typedef enum {')

	first = True
	for name, member in enum.items():
		if not first:
			print(',')
		first = False
		Indenter.printx(_constant_name(member.fullname()))
		value = getattr(member, 'value', None)
		alias = getattr(member, 'alias', None)
		if value is not None:
			val = " = %d" % value
		elif alias is not None:
			val, src, typ = alias
			if src:
				val = enums[src][val].fullname()
			elif typ == 'enum':
				val = enum[val].fullname()
			val = " = %s" % _constant_name(val)
		else:
			val = None
		if val is not None:
			sys.stdout.write(val)
	print('')

	Indenter.leave('} %s;' % _c_type(enum.name))
	print('')

def define_constant(c):
	Indenter.printline('#define %s %s' % (_constant_name(c.name), _c_literal(c.value)))
	print('')

def build():
	ipc = genipc.parse_xml('../src/ipc.xml')

	print(HEADER_PROLOG)
	print('')

	for n, c in ipc.constants.items():
		define_constant(c)
	print('')

	for n, e in ipc.enums.items():
		build_enum(e, ipc.enums)

	print('')
	print(HEADER_EPILOG)

if __name__ == '__main__':
	build()



# encoding: utf-8

from waflib.TaskGen import after_method,feature

UNKNOWN_OPTIONS_AS_ERRORS = '-Werror=unknown-warning-option'
VISIBILITY_FEATURE = '-fvisibility=hidden'

@feature('visibilityhidden')
@after_method('propagate_uselib_vars')
def add_visibility_hidden_flag(task):
    if task.env.have_visibility_cflag:
        task.env.append_unique('CFLAGS', VISIBILITY_FEATURE)
    if task.env.have_visibility_cxxflag:
        task.env.append_unique('CXXFLAGS', VISIBILITY_FEATURE)

def configure(ctx):
    have_cflag = False
    have_cxxflag = False

    ctx.env.stash()

    if ctx.check_cc(cflags=UNKNOWN_OPTIONS_AS_ERRORS, mandatory=False):
        ctx.env.append_unique('CFLAGS', UNKNOWN_OPTIONS_AS_ERRORS)

    if ctx.check_cxx(cxxflags=UNKNOWN_OPTIONS_AS_ERRORS, mandatory=False):
        ctx.env.append_unique('CXXFLAGS', UNKNOWN_OPTIONS_AS_ERRORS)

    if ctx.check_cc(cflags=VISIBILITY_FEATURE, mandatory=False):
        have_cflag = True
    if ctx.check_cxx(cflags=VISIBILITY_FEATURE, mandatory=False):
        have_cxxflag = True

    ctx.env.revert()

    ctx.env.have_visibility_cflag = have_cflag
    ctx.env.have_visibility_cxxflag = have_cxxflag



#!/usr/bin/env python
#  XMMS2 - X Music Multiplexer System
#  Copyright (C) 2003-2006 XMMS2 Team
# 
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#                   
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  This file is a part of the XMMS2 client tutorial #2
#  Here we will learn to retrieve results from a command

import xmmsclient
import os
import sys

"""
The first part of this program is
commented in tut1.py See that one for
instructions
"""

xmms = xmmsclient.XMMS("tutorial2")
try:
	xmms.connect(os.getenv("XMMS_PATH"))
except IOError, detail:
	print "Connection failed:", detail
	sys.exit(1)

"""
Now we send a command that will return
a result. Let's find out which entry
is currently playing. 

Note that this program has be run while 
xmms2 is playing something, otherwise
XMMS.playback_current_id will return 0.
"""
result = xmms.playback_current_id()

"""
We are still doing sync operations, wait for the
answer and block.
"""
result.wait()

"""
Also this time we need to check for errors.
Errors can occur on all commands, but not signals
and broadcasts. We will talk about these later.
"""
if result.iserror():
	print "playback current id returns error, %s" % result.get_error()

"""
Let's retrieve the value from the XMMSResult object.
You don't have to know what type of value is returned
in response to which command - simply call
XMMSResult.value()

In this case XMMS.playback_current_id will return a UINT
"""
id = result.value()

"""Print the value"""
print "Currently playing id is %d" % id




from waflib import Task, Errors
import os
import gzip

from waflib.TaskGen import feature,before_method

def gzip_func(task):
    infile = task.inputs[0].abspath()
    outfile = task.outputs[0].abspath()

    input = None
    outf = None
    output = None
    try:
        input = open(infile, 'rb')
        outf = open(outfile, 'wb')
        # Allow building deterministic manpages:
        #   https://reproducible-builds.org/docs/source-date-epoch/
        mtime = os.environ.get('SOURCE_DATE_EPOCH')
        if mtime:
            mtime = int(mtime)
        output = gzip.GzipFile(os.path.basename(infile), fileobj=outf, mode='wb', mtime=mtime)
        output.write(input.read())
    finally:
        if input:
            input.close()
        if output: # Must close before outf to flush compressed data.
            output.close()
        if outf:
            outf.close()

Task.task_factory('man', gzip_func, color='BLUE')

@feature('man')
@before_method('process_source')
def process_man(self):
    source = self.to_nodes(getattr(self, 'source', []))
    self.source = []

    section = getattr(self, 'section', None)

    for node in source:
        if not node:
            raise Errors.BuildError('cannot find input file %s for processing' % x)

        # s = section or node.name.rpartition('.')[2]
        s = section or "." in node.name and node.name.rsplit(".", 1)[1]
        if not s:
            raise Errors.BuildError('cannot determine man section from filename')

        out = self.path.find_or_declare(node.name + '.gz')

        tsk = self.create_task('man')
        tsk.set_inputs(node)
        tsk.set_outputs(out)

        self.bld.install_files('${MANDIR}/man%s' % s, out)


def configure(conf):
    return True



import os

def add_install_flag(bld, obj):
    env = bld.env

    if env['explicit_install_name']:
        libname = obj.env["cshlib_PATTERN"] % obj.target
        insname = os.path.join(obj.env["LIBDIR"], libname)
        obj.env.append_unique("LINKFLAGS", ["-install_name", insname])



from libc.stdint cimport int64_t

cdef extern from *:
	ctypedef char const_char "const char"
	ctypedef unsigned char const_uchar "const unsigned char"
	ctypedef int const_int "const int"

cdef extern from "xmmsc/xmmsc_idnumbers.h":
	ctypedef enum xmmsv_coll_type_t:
		XMMS_COLLECTION_TYPE_REFERENCE
		XMMS_COLLECTION_TYPE_UNIVERSE
		XMMS_COLLECTION_TYPE_UNION
		XMMS_COLLECTION_TYPE_INTERSECTION
		XMMS_COLLECTION_TYPE_COMPLEMENT
		XMMS_COLLECTION_TYPE_HAS
		XMMS_COLLECTION_TYPE_MATCH
		XMMS_COLLECTION_TYPE_TOKEN
		XMMS_COLLECTION_TYPE_EQUALS
		XMMS_COLLECTION_TYPE_NOTEQUAL
		XMMS_COLLECTION_TYPE_SMALLER
		XMMS_COLLECTION_TYPE_SMALLEREQ
		XMMS_COLLECTION_TYPE_GREATER
		XMMS_COLLECTION_TYPE_GREATEREQ
		XMMS_COLLECTION_TYPE_ORDER
		XMMS_COLLECTION_TYPE_LIMIT
		XMMS_COLLECTION_TYPE_MEDIASET
		XMMS_COLLECTION_TYPE_IDLIST

	ctypedef char *xmmsv_coll_namespace_t
	# XXX Trick cython compiler which doesn't deal well with extern
	# variables. Requires an explicit cast to <char *> when used directly.
	enum:
		XMMS_COLLECTION_NS_COLLECTIONS
		XMMS_COLLECTION_NS_PLAYLISTS
		XMMS_COLLECTION_NS_ALL

	# XXX Same trick. Requires an explicit cast <char *>
	enum:
		XMMS_ACTIVE_PLAYLIST

	const_char *xmmsv_default_source_pref[]

cdef extern from "xmmsc/xmmsv_c2c.h":
	int xmmsv_c2c_message_get_sender (xmmsv_t *c2c_msg)
	int xmmsv_c2c_message_get_id (xmmsv_t *c2c_msg)
	int xmmsv_c2c_message_get_destination (xmmsv_t *c2c_msg)
	xmmsv_t *xmmsv_c2c_message_get_payload (xmmsv_t *c2c_msg)

cdef extern from "xmmsc/xmmsv_service.h":
	xmmsv_t *xmmsv_sc_argument_new (const_char *name, const_char *docstring, xmmsv_type_t type, xmmsv_t *default_value)

	const_char *xmmsv_sc_argument_get_name (xmmsv_t *arg)
	const_char *xmmsv_sc_argument_get_docstring (xmmsv_t *arg)
	int xmmsv_sc_argument_get_type (xmmsv_t *arg)
	xmmsv_t *xmmsv_sc_argument_get_default_value (xmmsv_t *arg)


cdef extern from "xmmsc/xmmsv.h":

	# xmmsc/xmmsv_general.h

	ctypedef enum xmmsv_type_t:
		XMMSV_TYPE_NONE
		XMMSV_TYPE_ERROR
		XMMSV_TYPE_INT64
		XMMSV_TYPE_FLOAT
		XMMSV_TYPE_STRING
		XMMSV_TYPE_COLL
		XMMSV_TYPE_BIN
		XMMSV_TYPE_LIST
		XMMSV_TYPE_DICT
		XMMSV_TYPE_BITBUFFER

	ctypedef struct xmmsv_t

	xmmsv_t *xmmsv_new_none   ()
	xmmsv_t *xmmsv_new_error  (char *errstr)
	xmmsv_t *xmmsv_new_int    (int i)
	xmmsv_t *xmmsv_new_string (char *s)
	xmmsv_t *xmmsv_new_coll   (xmmsv_coll_type_t type)
	xmmsv_t *xmmsv_new_bin    (unsigned char *data, unsigned int len)

	xmmsv_t *xmmsv_copy (xmmsv_t *val)

	xmmsv_t *xmmsv_ref   (xmmsv_t *val)
	void     xmmsv_unref (xmmsv_t *val)

	xmmsv_type_t xmmsv_get_type (xmmsv_t *val)
	bint         xmmsv_is_type  (xmmsv_t *val, xmmsv_type_t t)

	bint xmmsv_get_error  (xmmsv_t *value, const_char **r)
	bint xmmsv_get_int64  (xmmsv_t *res, int64_t *r)
	bint xmmsv_get_float  (xmmsv_t *res, float *r)
	bint xmmsv_get_string (xmmsv_t *res, const_char **r)
	bint xmmsv_get_bin    (xmmsv_t *res, const_uchar **r, unsigned int *rlen)


	# xmmsc/xmmsv_list.h

	ctypedef int (*xmmsv_list_compare_func_t) (xmmsv_t **v1, xmmsv_t **v2)

	xmmsv_t *xmmsv_new_list ()

	bint xmmsv_list_get      (xmmsv_t *listv, int pos, xmmsv_t **val)
	bint xmmsv_list_set      (xmmsv_t *listv, int pos, xmmsv_t *val)
	bint xmmsv_list_append   (xmmsv_t *listv, xmmsv_t *val)
	bint xmmsv_list_insert   (xmmsv_t *listv, int pos, xmmsv_t *val)
	bint xmmsv_list_remove   (xmmsv_t *listv, int pos)
	bint xmmsv_list_move     (xmmsv_t *listv, int old_pos, int new_pos)
	bint xmmsv_list_clear    (xmmsv_t *listv)
	int  xmmsv_list_get_size (xmmsv_t *listv)
	bint xmmsv_list_restrict_type (xmmsv_t *listv, xmmsv_type_t type)
	bint xmmsv_list_has_type (xmmsv_t *listv, xmmsv_type_t type)

	bint xmmsv_list_get_string (xmmsv_t *v, int pos, const_char **val)
	bint xmmsv_list_get_int64  (xmmsv_t *v, int pos, int64_t *val)

	bint xmmsv_list_set_string (xmmsv_t *v, int pos, const_char *val)
	bint xmmsv_list_set_int    (xmmsv_t *v, int pos, int val)

	bint xmmsv_list_insert_string (xmmsv_t *v, int pos, const_char *val)
	bint xmmsv_list_insert_int    (xmmsv_t *v, int pos, int val)

	bint xmmsv_list_append_string (xmmsv_t *v, const_char *val)
	bint xmmsv_list_append_int    (xmmsv_t *v, int val)

	xmmsv_t *xmmsv_list_flatten (xmmsv_t *list, int depth)

	ctypedef void (*xmmsv_list_foreach_func) (xmmsv_t *value, void *user_data)
	bint xmmsv_list_foreach  (xmmsv_t *listv, xmmsv_list_foreach_func func, void* user_data)

	ctypedef struct xmmsv_list_iter_t
	bint xmmsv_get_list_iter (xmmsv_t *val, xmmsv_list_iter_t **it)
	void xmmsv_list_iter_explicit_destroy (xmmsv_list_iter_t *it)

	bint     xmmsv_list_iter_entry (xmmsv_list_iter_t *it, xmmsv_t **val)
	bint     xmmsv_list_iter_valid (xmmsv_list_iter_t *it)
	void     xmmsv_list_iter_first (xmmsv_list_iter_t *it)
	void     xmmsv_list_iter_last  (xmmsv_list_iter_t *it)
	void     xmmsv_list_iter_next  (xmmsv_list_iter_t *it)
	void     xmmsv_list_iter_prev  (xmmsv_list_iter_t *it)
	bint     xmmsv_list_iter_seek  (xmmsv_list_iter_t *it, int pos)
	bint     xmmsv_list_iter_tell  (xmmsv_list_iter_t *it)
	xmmsv_t *xmmsv_list_iter_get_parent (xmmsv_list_iter_t *it)

	bint xmmsv_list_iter_set (xmmsv_list_iter_t *it, xmmsv_t *val)
	bint xmmsv_list_iter_insert (xmmsv_list_iter_t *it, xmmsv_t *val)
	bint xmmsv_list_iter_remove (xmmsv_list_iter_t *it)

	bint xmmsv_list_iter_entry_string (xmmsv_list_iter_t *v, const_char **valval)
	bint xmmsv_list_iter_entry_int    (xmmsv_list_iter_t *v, int *val)

	bint xmmsv_list_iter_insert_string (xmmsv_list_iter_t *v, const_char *val)
	bint xmmsv_list_iter_insert_int    (xmmsv_list_iter_t *v, int val)


	# xmmsc/xmmsv_dict.h

	xmmsv_t *xmmsv_new_dict ()

	bint xmmsv_dict_get      (xmmsv_t *dictv, char *key, xmmsv_t **val)
	bint xmmsv_dict_set      (xmmsv_t *dictv, char *key, xmmsv_t *val)
	bint xmmsv_dict_remove   (xmmsv_t *dictv, char *key)
	bint xmmsv_dict_clear    (xmmsv_t *dictv)
	int  xmmsv_dict_get_size (xmmsv_t *dictv)
	bint xmmsv_dict_has_key  (xmmsv_t *dictv, const_char *key)

	bint xmmsv_dict_entry_get_string (xmmsv_t *val, const_char *key, const_char **r)
	bint xmmsv_dict_entry_get_int64    (xmmsv_t *val, const_char *key, int64_t *r)

	bint xmmsv_dict_set_string (xmmsv_t *val, const_char *key, const_char *el)
	bint xmmsv_dict_set_int    (xmmsv_t *val, const_char *key, int el)

	xmmsv_type_t xmmsv_dict_entry_get_type (xmmsv_t *val, const_char *key)

	ctypedef void (*xmmsv_dict_foreach_func) (char *key, xmmsv_t *value, void *user_data)
	bint xmmsv_dict_foreach  (xmmsv_t *dictv, xmmsv_dict_foreach_func func, void *user_data)

	ctypedef struct xmmsv_dict_iter_t
	bint xmmsv_get_dict_iter (xmmsv_t *val, xmmsv_dict_iter_t **it)
	void xmmsv_dict_iter_explicit_destroy (xmmsv_dict_iter_t *it)

	bint xmmsv_dict_iter_pair  (xmmsv_dict_iter_t *it, const_char **key, xmmsv_t **val)
	bint xmmsv_dict_iter_valid (xmmsv_dict_iter_t *it)
	void xmmsv_dict_iter_first (xmmsv_dict_iter_t *it)
	void xmmsv_dict_iter_next  (xmmsv_dict_iter_t *it)
	bint xmmsv_dict_iter_find  (xmmsv_dict_iter_t *it, const_char *key)

	bint xmmsv_dict_iter_set    (xmmsv_dict_iter_t *it, xmmsv_t *val)
	bint xmmsv_dict_iter_remove (xmmsv_dict_iter_t *it)

	bint xmmsv_dict_iter_pair_string (xmmsv_dict_iter_t *it, const_char **key, const_char **val)
	bint xmmsv_dict_iter_pair_int    (xmmsv_dict_iter_t *it, const_char **key, int *val)

	bint xmmsv_dict_iter_set_string (xmmsv_dict_iter_t *it, const_char *el)
	bint xmmsv_dict_iter_set_int    (xmmsv_dict_iter_t *it, int el)


	# xmmsc/xmmsv_coll.h

	void     xmmsv_coll_set_idlist     (xmmsv_t *coll, int *ids)
	void     xmmsv_coll_add_operand    (xmmsv_t *coll, xmmsv_t *op)
	void     xmmsv_coll_remove_operand (xmmsv_t *coll, xmmsv_t *op)
	xmmsv_t *xmmsv_coll_operands_get   (xmmsv_t *coll)

	bint xmmsv_coll_idlist_append    (xmmsv_t *coll, int id)
	bint xmmsv_coll_idlist_insert    (xmmsv_t *coll, int index, int id)
	bint xmmsv_coll_idlist_move      (xmmsv_t *coll, int index, int newindex)
	bint xmmsv_coll_idlist_remove    (xmmsv_t *coll, int index)
	bint xmmsv_coll_idlist_clear     (xmmsv_t *coll)
	bint xmmsv_coll_idlist_get_index (xmmsv_t *coll, int index, int *val)
	bint xmmsv_coll_idlist_set_index (xmmsv_t *coll, int index, int val)
	int  xmmsv_coll_idlist_get_size  (xmmsv_t *coll)

	bint               xmmsv_coll_is_type    (xmmsv_t *val, xmmsv_coll_type_t t)
	xmmsv_coll_type_t  xmmsv_coll_get_type   (xmmsv_t *coll)
	xmmsv_t           *xmmsv_coll_idlist_get (xmmsv_t *coll)

	void xmmsv_coll_attribute_set_string (xmmsv_t *coll, char *key, char *value)
	void xmmsv_coll_attribute_set_int    (xmmsv_t *coll, char *key, int value)
	void xmmsv_coll_attribute_set_value  (xmmsv_t *coll, char *key, xmmsv_t *value)

	bint xmmsv_coll_attribute_remove     (xmmsv_t *coll, char *key)

	bint xmmsv_coll_attribute_get_string (xmmsv_t *coll, char *key, const_char **value)
	bint xmmsv_coll_attribute_get_int64    (xmmsv_t *coll, char *key, int64_t *value)
	bint xmmsv_coll_attribute_get_value  (xmmsv_t *coll, char *key, xmmsv_t **value)

	xmmsv_t *xmmsv_coll_attributes_get (xmmsv_t *coll)
	void xmmsv_coll_attributes_set (xmmsv_t *coll, xmmsv_t *attributes)

	xmmsv_t *xmmsv_coll_universe ()

	xmmsv_t *xmmsv_coll_add_order_operator  (xmmsv_t *coll, xmmsv_t *order)
	xmmsv_t *xmmsv_coll_add_order_operators (xmmsv_t *coll, xmmsv_t *order)
	xmmsv_t *xmmsv_coll_add_limit_operator  (xmmsv_t *coll, int lim_start, int lim_len)


	# xmmsc/xmmsv_bitbuffer.h

	xmmsv_t *xmmsv_new_bitbuffer_ro (const_uchar *v, int len)
	xmmsv_t *xmmsv_new_bitbuffer ()

	bint xmmsv_bitbuffer_get_bits    (xmmsv_t *v, int bits, int *res)
	bint xmmsv_bitbuffer_get_data    (xmmsv_t *v, char *b, int len)
	bint xmmsv_bitbuffer_put_bits    (xmmsv_t *v, int bits, int d)
	bint xmmsv_bitbuffer_put_bits_at (xmmsv_t *v, intd, int offset)
	bint xmmsv_bitbuffer_put_data    (xmmsv_t *v, const_uchar *b, int len)

	bint xmmsv_bitbuffer_align    (xmmsv_t *v)
	bint xmmsv_bitbuffer_goto     (xmmsv_t *v, int pos)
	int  xmmsv_bitbuffer_pos      (xmmsv_t *v)
	bint xmmsv_bitbuffer_rewind   (xmmsv_t *v)
	bint xmmsv_bitbuffer_end      (xmmsv_t *v)
	int  xmmsv_bitbuffer_len      (xmmsv_t *v)

	const_uchar * xmmsv_bitbuffer_buffer (xmmsv_t *v)
	bint          xmmsv_get_bitbuffer    (xmmsv_t *val, const_uchar **r, int *rlen)

	bint xmmsv_bitbuffer_serialize_value   (xmmsv_t *bb, xmmsv_t *v)
	bint xmmsv_bitbuffer_deserialize_value (xmmsv_t *bb, xmmsv_t **val)


	# xmmsc/xmmsv_util.h

	xmmsv_t *xmmsv_decode_url (xmmsv_t *url)

	int xmmsv_utf8_validate (const_char *s)
	xmmsv_t *xmmsv_propdict_to_dict (xmmsv_t *propdict, const_char **src_pref)

	int xmmsv_dict_format (char *target, int len, const_char *fmt, xmmsv_t *val)

	xmmsv_t *xmmsv_serialize (xmmsv_t *v)
	xmmsv_t *xmmsv_deserialize (xmmsv_t *v)


cdef extern from "xmmsclient/xmmsclient.h":
	bint xmmsv_coll_parse (char *pattern, xmmsv_t **coll)



from waflib import TaskGen
import sys

TaskGen.declare_chain(
		name='genpy',
		rule='${PYTHON} ${SRC} ${GENPY_ARGS} > ${TGT}',
		reentrant=True,
		color='BLUE',
		ext_in='.genpy',
		ext_out='.c',
		before='c')

def configure(conf):
    conf.env['PYTHON'] = sys.executable
    conf.env['GENPY_EXT'] = ['.genpy']
    conf.env['GENPY_ARGS'] = ''
    return True



import os
try: from hashlib import sha1 as sha
except ImportError: from sha import sha

def gitsha(path):
    h = sha()
    data = file(path, 'rb').read()
    h.update("blob %d\0" % len(data))
    h.update(data)
    return h.hexdigest()

def git_info():
    commithash = os.popen('git rev-parse --verify HEAD 2>/dev/null').read().strip()
    if not commithash:
        raise ValueError("Couldn't get hash")
    if os.getuid() == os.stat(".git/index").st_uid:
        os.system('git update-index --refresh >/dev/null')
    else:
        print("NOT updating git cache, local changes might not be detected")
    changed = bool(os.popen('git diff-index -r HEAD').read())
    return commithash[:8], changed

def snapshot_info():
    info = file('commithash').read().split('\n')

    commithash = info[0]

    changed = False
    for line in [a for a in info[2:] if a]:
        [mode, tag, sha, path] = line.split(None, 4)
        if tag != 'blob':
            continue
        if gitsha(path) != sha:
            changed = True
            break
    return commithash, changed

def get_info():
    try:
        return git_info()
    except:
        try:
            return snapshot_info()
        except:
            return 'Unknown', False

def get_info_str():
    commithash, changed = get_info()
    if changed:
        changed = " + local changes"
    else:
        changed = ""

    return "%s%s" % (commithash, changed)

submodule_status = {'-':'missing', '+':'outdated', ' ':'uptodate'}
def git_submodules():
    submodules = {}
    for l in os.popen('git submodule').read().split('\n'):
        status = submodule_status.get(l and l[0] or "", 'uptodate')
        l = l[1:].strip()
        if not l:
            continue
        commithash, folder = l.strip().split()[:2]
        submodules[folder] = (status, commithash[:8])
    return submodules

def get_submodules():
    try:
        return git_submodules()
    except:
        return {}



from cxmmsvalue cimport *

cdef extern from "xmmsc/xmmsc_idnumbers.h":
	ctypedef enum xmms_playback_status_t:
		XMMS_PLAYBACK_STATUS_STOP
		XMMS_PLAYBACK_STATUS_PLAY
		XMMS_PLAYBACK_STATUS_PAUSE

	ctypedef enum xmms_playlist_changed_action_t:
		XMMS_PLAYLIST_CHANGED_ADD
		XMMS_PLAYLIST_CHANGED_INSERT
		XMMS_PLAYLIST_CHANGED_SHUFFLE
		XMMS_PLAYLIST_CHANGED_REMOVE
		XMMS_PLAYLIST_CHANGED_CLEAR
		XMMS_PLAYLIST_CHANGED_MOVE
		XMMS_PLAYLIST_CHANGED_SORT
		XMMS_PLAYLIST_CHANGED_UPDATE

	ctypedef enum xmms_plugin_type_t:
		XMMS_PLUGIN_TYPE_ALL
		XMMS_PLUGIN_TYPE_XFORM
		XMMS_PLUGIN_TYPE_OUTPUT

	ctypedef enum xmmsc_medialib_entry_status_t:
		XMMS_MEDIALIB_ENTRY_STATUS_NEW
		XMMS_MEDIALIB_ENTRY_STATUS_OK
		XMMS_MEDIALIB_ENTRY_STATUS_RESOLVING
		XMMS_MEDIALIB_ENTRY_STATUS_NOT_AVAILABLE
		XMMS_MEDIALIB_ENTRY_STATUS_REHASH

	ctypedef enum xmms_collection_changed_action_t:
		XMMS_COLLECTION_CHANGED_ADD
		XMMS_COLLECTION_CHANGED_UPDATE
		XMMS_COLLECTION_CHANGED_RENAME
		XMMS_COLLECTION_CHANGED_REMOVE

	ctypedef enum xmms_playback_seek_mode_t:
		XMMS_PLAYBACK_SEEK_CUR
		XMMS_PLAYBACK_SEEK_SET

	ctypedef enum xmmsc_c2c_reply_policy_t:
		XMMS_C2C_REPLY_POLICY_NO_REPLY
		XMMS_C2C_REPLY_POLICY_SINGLE_REPLY
		XMMS_C2C_REPLY_POLICY_MULTI_REPLY

cdef extern from "xmmsc/xmmsc_util.h":
	cdef enum:
		XMMS_PATH_MAX

cdef extern from "xmmsc/xmmsv.h":
	ctypedef struct xmmsv_t

cdef extern from "xmmsc/xmmsc_visualization.h":
	cdef enum:
		XMMSC_VISUALIZATION_WINDOW_SIZE

cdef extern from "xmmsclient/xmmsclient.h":
	ctypedef struct xmmsc_connection_t
	ctypedef struct xmmsc_result_t

	ctypedef enum xmmsc_result_type_t:
		XMMSC_RESULT_CLASS_DEFAULT
		XMMSC_RESULT_CLASS_SIGNAL
		XMMSC_RESULT_CLASS_BROADCAST

	ctypedef void (*xmmsc_disconnect_func_t)           (void *user_data)
	ctypedef void (*xmmsc_user_data_free_func_t)       (void *user_data)
	ctypedef void (*xmmsc_io_need_out_callback_func_t) (int, void*)

	xmmsc_connection_t *xmmsc_init    (char *clientname)
	bint                xmmsc_connect (xmmsc_connection_t *c, char *clientname)
	xmmsc_connection_t *xmmsc_ref     (xmmsc_connection_t *c)
	void                xmmsc_unref   (xmmsc_connection_t *c)

	void xmmsc_lock_set (xmmsc_connection_t *c, void *lock, void (*lockfunc)(void *), void (*unlockfunc)(void *))
	void xmmsc_disconnect_callback_set (xmmsc_connection_t *c, xmmsc_disconnect_func_t callback, void *userdata)
	void xmmsc_disconnect_callback_set_full (xmmsc_connection_t *c, xmmsc_disconnect_func_t disconnect_func, void *userdata, xmmsc_user_data_free_func_t free_func)

	void xmmsc_io_need_out_callback_set      (xmmsc_connection_t *c, xmmsc_io_need_out_callback_func_t callback, void *userdata)
	void xmmsc_io_need_out_callback_set_full (xmmsc_connection_t *c, xmmsc_io_need_out_callback_func_t callback, void *userdata, xmmsc_user_data_free_func_t free_func)

	void xmmsc_io_disconnect (xmmsc_connection_t *c)
	bint xmmsc_io_want_out   (xmmsc_connection_t *c)
	bint xmmsc_io_out_handle (xmmsc_connection_t *c)
	bint xmmsc_io_in_handle  (xmmsc_connection_t *c)
	int  xmmsc_io_fd_get     (xmmsc_connection_t *c)

	char *xmmsc_get_last_error (xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_quit(xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_broadcast_quit (xmmsc_connection_t *c)

	char *xmmsc_userconfdir_get (char *buf, int len)

	# Playlists
	xmmsc_result_t *xmmsc_playlist_list    (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_playlist_create  (xmmsc_connection_t *c, char *playlist)
	xmmsc_result_t *xmmsc_playlist_shuffle (xmmsc_connection_t *c, char *playlist)

	xmmsc_result_t *xmmsc_playlist_add_full    (xmmsc_connection_t *c, char *playlist, char *, xmmsv_t *)
	xmmsc_result_t *xmmsc_playlist_add_url     (xmmsc_connection_t *c, char *playlist, char *)
	xmmsc_result_t *xmmsc_playlist_add_id      (xmmsc_connection_t *c, char *playlist, int)
	xmmsc_result_t *xmmsc_playlist_add_encoded (xmmsc_connection_t *c, char *, char *)
	xmmsc_result_t *xmmsc_playlist_add_idlist  (xmmsc_connection_t *c, char *playlist, xmmsv_t *coll)
	xmmsc_result_t *xmmsc_playlist_add_collection (xmmsc_connection_t *c, char *playlist, xmmsv_t *coll, xmmsv_t *order)

	xmmsc_result_t *xmmsc_playlist_remove_entry (xmmsc_connection_t *c, char *playlist, int)
	xmmsc_result_t *xmmsc_playlist_clear        (xmmsc_connection_t *c, char *playlist)
	xmmsc_result_t *xmmsc_playlist_remove       (xmmsc_connection_t *c, char *playlist)
	xmmsc_result_t *xmmsc_playlist_list_entries (xmmsc_connection_t *c, char *playlist)
	xmmsc_result_t *xmmsc_playlist_sort         (xmmsc_connection_t *c, char *playlist, xmmsv_t *properties)
	xmmsc_result_t *xmmsc_playlist_set_next     (xmmsc_connection_t *c, int pos)
	xmmsc_result_t *xmmsc_playlist_set_next_rel (xmmsc_connection_t *c, int rpos)
	xmmsc_result_t *xmmsc_playlist_move_entry   (xmmsc_connection_t *c, char *playlist, int id, int movement)
	xmmsc_result_t *xmmsc_playlist_current_pos  (xmmsc_connection_t *c, char *playlist)
	xmmsc_result_t *xmmsc_playlist_current_active (xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_playlist_insert_full (xmmsc_connection_t *c, char *playlist, int pos, char *url, xmmsv_t *args)
	xmmsc_result_t *xmmsc_playlist_insert_url  (xmmsc_connection_t *c, char *playlist, int pos, char *url)
	xmmsc_result_t *xmmsc_playlist_insert_id   (xmmsc_connection_t *c, char *playlist, int pos, int)
	xmmsc_result_t *xmmsc_playlist_insert_encoded (xmmsc_connection_t *c, char *playlist, int pos, char *url)
	xmmsc_result_t *xmmsc_playlist_insert_collection (xmmsc_connection_t *c, char *playlist, int pos, xmmsv_t *coll, xmmsv_t *order)

	xmmsc_result_t *xmmsc_playlist_load (xmmsc_connection_t *c, char *playlist)
	xmmsc_result_t *xmmsc_playlist_radd (xmmsc_connection_t *c, char *playlist, char *path)
	xmmsc_result_t *xmmsc_playlist_radd_encoded (xmmsc_connection_t *c, char *playlist, char *path)
	xmmsc_result_t *xmmsc_playlist_rinsert (xmmsc_connection_t *c, char *playlist, int pos, char *url)
	xmmsc_result_t *xmmsc_playlist_rinsert_encoded (xmmsc_connection_t *c, char *playlist, int pos, char *url)

	xmmsc_result_t *xmmsc_broadcast_playlist_changed     (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_broadcast_playlist_current_pos (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_broadcast_playlist_loaded      (xmmsc_connection_t *c)

	# Playback
	xmmsc_result_t *xmmsc_playback_stop         (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_playback_tickle       (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_playback_start        (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_playback_pause        (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_playback_current_id   (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_playback_seek_ms      (xmmsc_connection_t *c, int milliseconds, xmms_playback_seek_mode_t whence)
	xmmsc_result_t *xmmsc_playback_seek_samples (xmmsc_connection_t *c, int samples, xmms_playback_seek_mode_t whence)
	xmmsc_result_t *xmmsc_playback_playtime     (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_playback_status       (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_playback_volume_set   (xmmsc_connection_t *c, char *channel, int volume)
	xmmsc_result_t *xmmsc_playback_volume_get   (xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_broadcast_playback_volume_changed (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_broadcast_playback_status         (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_broadcast_playback_current_id     (xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_signal_playback_playtime(xmmsc_connection_t *c)

	# Config
	xmmsc_result_t *xmmsc_config_set_value      (xmmsc_connection_t *c, char *key, char *val)
	xmmsc_result_t *xmmsc_config_list_values    (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_config_get_value      (xmmsc_connection_t *c, char *key)
	xmmsc_result_t *xmmsc_config_register_value (xmmsc_connection_t *c, char *valuename, char *defaultvalue)

	xmmsc_result_t *xmmsc_broadcast_config_value_changed (xmmsc_connection_t *c)

	# Stats
	xmmsc_result_t *xmmsc_main_list_plugins (xmmsc_connection_t *c, xmms_plugin_type_t type)
	xmmsc_result_t *xmmsc_main_stats        (xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_broadcast_mediainfo_reader_status (xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_signal_mediainfo_reader_unindexed (xmmsc_connection_t *c)

	# Visualization
	xmmsc_result_t *xmmsc_visualization_version      (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_visualization_init         (xmmsc_connection_t *c)
	int             xmmsc_visualization_init_handle  (xmmsc_result_t *res)
	xmmsc_result_t *xmmsc_visualization_start        (xmmsc_connection_t *c, int vv)
	void            xmmsc_visualization_start_handle (xmmsc_connection_t *c, xmmsc_result_t *res)
	bint            xmmsc_visualization_started      (xmmsc_connection_t *c, int vv)
	bint            xmmsc_visualization_errored      (xmmsc_connection_t *c, int vv)

	xmmsc_result_t *xmmsc_visualization_property_set   (xmmsc_connection_t *c, int v, char *key, char *value)
	xmmsc_result_t *xmmsc_visualization_properties_set (xmmsc_connection_t *c, int v, xmmsv_t *props)

	int  xmmsc_visualization_chunk_get (xmmsc_connection_t *c, int vv, short *buffer, int drawtime, unsigned int blocking)
	void xmmsc_visualization_shutdown (xmmsc_connection_t *c, int v)

	# Medialib
	xmmsc_result_t *xmmsc_medialib_add_entry           (xmmsc_connection_t *c, char *url)
	xmmsc_result_t *xmmsc_medialib_add_entry_full      (xmmsc_connection_t *c, char *url, xmmsv_t *args)
	xmmsc_result_t *xmmsc_medialib_add_entry_encoded   (xmmsc_connection_t *c, char *url)
	xmmsc_result_t *xmmsc_medialib_get_info            (xmmsc_connection_t *c, int id)
	xmmsc_result_t *xmmsc_medialib_import_path         (xmmsc_connection_t *c, char *path)
	xmmsc_result_t *xmmsc_medialib_import_path_encoded (xmmsc_connection_t *c, char *path)
	xmmsc_result_t *xmmsc_medialib_rehash              (xmmsc_connection_t *c, unsigned int)
	xmmsc_result_t *xmmsc_medialib_get_id              (xmmsc_connection_t *c, char *url)
	xmmsc_result_t *xmmsc_medialib_get_id_encoded      (xmmsc_connection_t *c, char *url)
	xmmsc_result_t *xmmsc_medialib_remove_entry        (xmmsc_connection_t *conn, int id)
	xmmsc_result_t *xmmsc_medialib_move_entry          (xmmsc_connection_t *conn, int id, char *url)

	xmmsc_result_t *xmmsc_medialib_entry_property_set_int (xmmsc_connection_t *c, int id, char *key, int value)
	xmmsc_result_t *xmmsc_medialib_entry_property_set_int_with_source (xmmsc_connection_t *c, int id, char *source, char *key, int value)

	xmmsc_result_t *xmmsc_medialib_entry_property_set_str (xmmsc_connection_t *c, int id, char *key, char *value)
	xmmsc_result_t *xmmsc_medialib_entry_property_set_str_with_source (xmmsc_connection_t *c, int id, char *source, char *key, char *value)

	xmmsc_result_t *xmmsc_medialib_entry_property_remove (xmmsc_connection_t *c, int id, char *key)
	xmmsc_result_t *xmmsc_medialib_entry_property_remove_with_source (xmmsc_connection_t *c, int id, char *source, char *key)

	xmmsc_result_t *xmmsc_xform_media_browse         (xmmsc_connection_t *c, char *url)
	xmmsc_result_t *xmmsc_xform_media_browse_encoded (xmmsc_connection_t *c, char *url)

	xmmsc_result_t *xmmsc_bindata_add      (xmmsc_connection_t *c, unsigned char *data, int len)
	xmmsc_result_t *xmmsc_bindata_retrieve (xmmsc_connection_t *c, char *hash)
	xmmsc_result_t *xmmsc_bindata_remove   (xmmsc_connection_t *c, char *hash)
	xmmsc_result_t *xmmsc_bindata_list     (xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_broadcast_medialib_entry_added   (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_broadcast_medialib_entry_updated (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_broadcast_medialib_entry_removed (xmmsc_connection_t *c)

	# Collections
	xmmsc_result_t *xmmsc_coll_get    (xmmsc_connection_t *c, char *collname, xmmsv_coll_namespace_t ns)
	xmmsc_result_t *xmmsc_coll_list   (xmmsc_connection_t *c, xmmsv_coll_namespace_t ns)
	xmmsc_result_t *xmmsc_coll_save   (xmmsc_connection_t *c, xmmsv_t *coll, char* name, xmmsv_coll_namespace_t ns)
	xmmsc_result_t *xmmsc_coll_remove (xmmsc_connection_t *c, char* name, xmmsv_coll_namespace_t ns)
	xmmsc_result_t *xmmsc_coll_find   (xmmsc_connection_t *c, int mediaid, xmmsv_coll_namespace_t ns)
	xmmsc_result_t *xmmsc_coll_rename (xmmsc_connection_t *c, char* from_name, char* to_name, xmmsv_coll_namespace_t ns)
	xmmsc_result_t *xmmsc_coll_idlist_from_playlist_file (xmmsc_connection_t *c, char *path)
	xmmsc_result_t *xmmsc_coll_sync   (xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_coll_query       (xmmsc_connection_t *c, xmmsv_t *coll, xmmsv_t *fetch)
	xmmsc_result_t *xmmsc_coll_query_ids   (xmmsc_connection_t *c, xmmsv_t *coll, xmmsv_t *order, unsigned int limit_start, unsigned int limit_len)
	xmmsc_result_t *xmmsc_coll_query_infos (xmmsc_connection_t *c, xmmsv_t *coll, xmmsv_t *order, unsigned int limit_start, unsigned int limit_len,  xmmsv_t *fetch, xmmsv_t *group)

	xmmsc_result_t *xmmsc_broadcast_collection_changed (xmmsc_connection_t *c)


	# C2C
	xmmsc_result_t *xmmsc_c2c_send  (xmmsc_connection_t *c, int dest, xmmsc_c2c_reply_policy_t reply_policy, xmmsv_t *payload)
	xmmsc_result_t *xmmsc_c2c_reply (xmmsc_connection_t *c, int msgid, xmmsc_c2c_reply_policy_t reply_policy, xmmsv_t *payload)
	int xmmsc_c2c_get_own_id (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_c2c_get_connected_clients (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_c2c_ready (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_c2c_get_ready_clients (xmmsc_connection_t *c)

	xmmsc_result_t *xmmsc_broadcast_c2c_message (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_broadcast_c2c_ready (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_broadcast_c2c_client_connected (xmmsc_connection_t *c)
	xmmsc_result_t *xmmsc_broadcast_c2c_client_disconnected (xmmsc_connection_t *c)


	# Service
	ctypedef enum xmmsc_sc_command_t:
		XMMSC_SC_CALL
		XMMSC_SC_BROADCAST_SUBSCRIBE
		XMMSC_SC_INTROSPECT

	# XXX Requires an explicit cast to <char *> when used
	enum:
		XMMSC_SC_CMD_KEY
		XMMSC_SC_ARGS_KEY
		XMMSC_SC_CALL_METHOD_KEY
		XMMSC_SC_CALL_PARGS_KEY
		XMMSC_SC_CALL_NARGS_KEY
		XMMSC_SC_INSTROSPECT_PATH_KEY
		XMMSC_SC_INSTROSPECT_TYPE_KEY
		XMMSC_SC_INSTROSPECT_KEYFILTER_KEY

	ctypedef struct xmmsc_sc_namespace_t
	ctypedef xmmsv_t *(*xmmsc_sc_method_t) (xmmsv_t *pargs, xmmsv_t *nargs, void *udata)

	xmmsc_sc_namespace_t *xmmsc_sc_init (xmmsc_connection_t *c)

	xmmsc_sc_namespace_t *xmmsc_sc_namespace_root (xmmsc_connection_t *c)
	xmmsc_sc_namespace_t *xmmsc_sc_namespace_lookup (xmmsc_connection_t *c, xmmsv_t *nms)
	xmmsc_sc_namespace_t *xmmsc_sc_namespace_new (xmmsc_sc_namespace_t *parent, const_char *name, const_char *docstring)
	xmmsc_sc_namespace_t *xmmsc_sc_namespace_get (xmmsc_sc_namespace_t *parent, const_char *name)
	bint xmmsc_sc_namespace_add_constant (xmmsc_sc_namespace_t *nms, const_char *key, xmmsv_t *value)
	void xmmsc_sc_namespace_remove_constant (xmmsc_sc_namespace_t *nms, const char *key)

	bint xmmsc_sc_namespace_add_method (xmmsc_sc_namespace_t *nmd, xmmsc_sc_method_t method, const_char *name, const_char *docstring, xmmsv_t *pargs, xmmsv_t *nargs, bint va_pos, bint va_named, void *udata)
	bint xmmsc_sc_namespace_add_method_noarg (xmmsc_sc_namespace_t *nms, xmmsc_sc_method_t method, const_char *name, const_char *docstring, void * udata)

	bint xmmsc_sc_namespace_add_broadcast (xmmsc_sc_namespace_t *nms, const_char *name, const_char *docstring)

	void xmmsc_sc_namespace_remove (xmmsc_sc_namespace_t *nms, xmmsv_t *path)
	bint xmmsc_sc_broadcast_emit (xmmsc_connection_t *c, xmmsv_t *broadcast, xmmsv_t *value)
	xmmsc_result_t *xmmsc_sc_broadcast_subscribe (xmmsc_connection_t *c, int dest, xmmsv_t *broadcast)

	xmmsc_result_t *xmmsc_sc_call (xmmsc_connection_t *c, int dest, xmmsv_t *method, xmmsv_t *pargs, xmmsv_t *nargs)

	xmmsc_result_t *xmmsc_sc_introspect_namespace (xmmsc_connection_t *c, int dest, xmmsv_t *nms)
	xmmsc_result_t *xmmsc_sc_introspect_method (xmmsc_connection_t *c, int dest, xmmsv_t *method)
	xmmsc_result_t *xmmsc_sc_introspect_broadcast (xmmsc_connection_t *c, int dest, xmmsv_t *broadcast)
	xmmsc_result_t *xmmsc_sc_introspect_constant (xmmsc_connection_t *c, int dest, xmmsv_t *nms, const_char *key)
	xmmsc_result_t *xmmsc_sc_introspect_docstring (xmmsc_connection_t *c, int dest, xmmsv_t *path)


	# Results
	ctypedef bint (*xmmsc_result_notifier_t) (xmmsv_t *val, void *user_data)

	xmmsc_result_type_t xmmsc_result_get_class  (xmmsc_result_t *res)
	void                xmmsc_result_disconnect (xmmsc_result_t *res)

	xmmsc_result_t *xmmsc_result_ref   (xmmsc_result_t *res)
	void            xmmsc_result_unref (xmmsc_result_t *res)

	void xmmsc_result_notifier_set      (xmmsc_result_t *res, xmmsc_result_notifier_t func, void *user_data)
	void xmmsc_result_notifier_set_full (xmmsc_result_t *res, xmmsc_result_notifier_t func, void *user_data, xmmsc_user_data_free_func_t free_func)
	void xmmsc_result_wait              (xmmsc_result_t *res)

	xmmsv_t *xmmsc_result_get_value (xmmsc_result_t *res)
